#! /usr/bin/python3

# Copyright (C) 2013 Canonical Ltd.
# Author: Brian Murray <brian@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Use and manage a Click chroot."""

from __future__ import print_function

from argparse import ArgumentParser, REMAINDER
from contextlib import contextmanager
import os

from click_package.chroot import (
    ClickChroot,
    ClickChrootAlreadyExistsException,
    ClickChrootDoesNotExistException,
)
from click_package import osextras


def requires_root(parser):
    if os.getuid() != 0:
        parser.error("must be run as root; try sudo")


@contextmanager
def message_on_error(exc, msg):
    """
    Context Manager that prints the error message 'msg' on exception 'exc'
    """
    try:
        yield
    except exc:
        print(msg)


# FIXME: i18n(?)
class ErrorMessages:
    EXISTS = """A chroot for that name and architecture already exists.
Please see the man-page how to use it."""
    NOT_EXISTS = """A chroot for that name and architecture does not exist.
Please use 'create' to create it."""


def create(parser, args):
    if not osextras.find_on_path("debootstrap"):
        parser.error(
            "debootstrap not installed and configured; install click-dev and "
            "debootstrap")
    requires_root(parser)
    chroot = ClickChroot(
        args.architecture, args.framework, name=args.name, series=args.series)
    with message_on_error(
            ClickChrootAlreadyExistsException, ErrorMessages.EXISTS):
        return chroot.create(args.keep_broken_chroot)
    # if we reach this point there was a error so return exit_status 1
    return 1


def install(parser, args):
    packages = args.packages
    chroot = ClickChroot(
        args.architecture, args.framework, name=args.name,
        session=args.session)
    with message_on_error(
            ClickChrootDoesNotExistException, ErrorMessages.NOT_EXISTS):
        return chroot.install(*packages)
    # if we reach this point there was a error so return exit_status 1
    return 1


def destroy(parser, args):
    requires_root(parser)
    # ask for confirmation?
    chroot = ClickChroot(args.architecture, args.framework, name=args.name)
    with message_on_error(
            ClickChrootDoesNotExistException, ErrorMessages.NOT_EXISTS):
        return chroot.destroy()
    # if we reach this point there was a error so return exit_status 1
    return 1


def execute(parser, args):
    program = args.program
    if not program:
        program = ["/bin/bash"]
    chroot = ClickChroot(
        args.architecture, args.framework, name=args.name,
        session=args.session)
    with message_on_error(
            ClickChrootDoesNotExistException, ErrorMessages.NOT_EXISTS):
        return chroot.run(*program)
    # if we reach this point there was a error so return exit_status 1
    return 1


def maint(parser, args):
    program = args.program
    if not program:
        program = ["/bin/bash"]
    chroot = ClickChroot(
        args.architecture, args.framework, name=args.name,
        session=args.session)
    with message_on_error(
            ClickChrootDoesNotExistException, ErrorMessages.NOT_EXISTS):
        return chroot.maint(*program)
    # if we reach this point there was a error so return exit_status 1
    return 1


def upgrade(parser, args):
    chroot = ClickChroot(
        args.architecture, args.framework, name=args.name,
        session=args.session)
    with message_on_error(
            ClickChrootDoesNotExistException, ErrorMessages.NOT_EXISTS):
        return chroot.upgrade()
    # if we reach this point there was a error so return exit_status 1
    return 1


def begin_session(parser, args):
    chroot = ClickChroot(
        args.architecture, args.framework, name=args.name,
        session=args.session)
    with message_on_error(
            ClickChrootDoesNotExistException, ErrorMessages.NOT_EXISTS):
        return chroot.begin_session()
    # if we reach this point there was a error so return exit_status 1
    return 1


def end_session(parser, args):
    chroot = ClickChroot(
        args.architecture, args.framework, name=args.name,
        session=args.session)
    with message_on_error(
            ClickChrootDoesNotExistException, ErrorMessages.NOT_EXISTS):
        return chroot.end_session()
    # if we reach this point there was a error so return exit_status 1
    return 1


def exists(parser, args):
    chroot = ClickChroot(args.architecture, args.framework, name=args.name)
    # return shell exit codes 0 on success, 1 on failure
    if chroot.exists():
        return 0
    else:
        return 1


def run(argv):
    parser = ArgumentParser("click chroot")
    subparsers = parser.add_subparsers(
        description="management subcommands",
        help="valid commands")
    parser.add_argument(
        "-a", "--architecture", required=True,
        help="architecture for the chroot")
    parser.add_argument(
        "-f", "--framework", default="ubuntu-sdk-14.04",
        help="framework for the chroot (default: ubuntu-sdk-14.04)")
    parser.add_argument(
        "-s", "--series",
        help="series to use for a newly-created chroot (defaults to a series "
             "appropriate for the framework)")
    parser.add_argument(
        "-n", "--name", default="click",
        help=(
            "name of the chroot (default: click; the framework and "
            "architecture will be appended)"))
    create_parser = subparsers.add_parser(
        "create",
        help="create a chroot of the provided architecture")
    create_parser.add_argument(
        "-k", "--keep-broken-chroot", default=False, action="store_true",
        help="Keep the chroot even if creating it fails (default is to delete "
              "it)")
    create_parser.set_defaults(func=create)
    destroy_parser = subparsers.add_parser(
        "destroy",
        help="destroy the chroot")
    destroy_parser.set_defaults(func=destroy)
    upgrade_parser = subparsers.add_parser(
        "upgrade",
        help="upgrade the chroot")
    upgrade_parser.add_argument(
        "-n", "--session-name",
        dest='session',
        help="persistent chroot session name to upgrade")
    upgrade_parser.set_defaults(func=upgrade)
    install_parser = subparsers.add_parser(
        "install",
        help="install packages in the chroot")
    install_parser.add_argument(
        "-n", "--session-name",
        dest='session',
        help="persistent chroot session name to install packages in")
    install_parser.add_argument(
        "packages", nargs="+",
        help="packages to install")
    install_parser.set_defaults(func=install)
    execute_parser = subparsers.add_parser(
        "run",
        help="run a program in the chroot")
    execute_parser.add_argument(
        "-n", "--session-name",
        dest='session',
        help="persistent chroot session name to run a program in")
    execute_parser.add_argument(
        "program", nargs=REMAINDER,
        help="program to run with arguments")
    execute_parser.set_defaults(func=execute)
    maint_parser = subparsers.add_parser(
        "maint",
        help="run a maintenance command in the chroot")
    maint_parser.add_argument(
        "-n", "--session-name",
        dest='session',
        help="persistent chroot session name to run a maintenance command in")
    maint_parser.add_argument(
        "program", nargs=REMAINDER,
        help="program to run with arguments")
    maint_parser.set_defaults(func=maint)
    begin_parser = subparsers.add_parser(
        "begin-session",
        help="begin a persistent chroot session")
    begin_parser.add_argument(
        "session",
        help="new session name")
    begin_parser.set_defaults(func=begin_session)
    end_parser = subparsers.add_parser(
        "end-session",
        help="end a persistent chroot session")
    end_parser.add_argument(
        "session",
        help="session name to end")
    end_parser.set_defaults(func=end_session)
    exists_parser = subparsers.add_parser(
        "exists",
        help="test if the given chroot exists")
    exists_parser.set_defaults(func=exists)
    args = parser.parse_args(argv)
    if not hasattr(args, "func"):
        parser.print_help()
        return 1
    if (not osextras.find_on_path("schroot") or
            not os.path.exists("/etc/schroot/click/fstab")):
        parser.error(
            "schroot not installed and configured; install click-dev and "
            "schroot")
    return args.func(parser, args)



# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt <michael.vogt@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Integration tests for the click CLI info command."""

import json
import os
import subprocess

from .helpers import ClickTestCase


class TestInfo(ClickTestCase):
    def test_info_from_path(self):
        name = "com.example.foo"
        path_to_click = self._make_click(name)
        output = subprocess.check_output([
            self.click_binary, "info", path_to_click], universal_newlines=True)
        self.assertEqual(name, json.loads(output)["name"])

    def test_info_installed_click(self):
        name = "com.example.foo"
        user = os.environ.get("USER", "root")
        path_to_click = self._make_click(name, framework="")
        self.click_install(path_to_click, name, user)
        output = subprocess.check_output([
            self.click_binary, "info", name], universal_newlines=True)
        self.assertEqual(json.loads(output)["name"], name)

    def test_info_file_in_package(self):
        name = "org.example.info"
        version = "1.0"
        click_pkg = self._make_click(name=name, version=version, framework="")
        subprocess.check_call(
            [self.click_binary, "install", "--allow-unauthenticated",
             "--all-users", click_pkg])
        self.addCleanup(
            subprocess.check_call,
            [self.click_binary, "unregister", "--all-users", name])
        output = subprocess.check_output(
            [self.click_binary, "info",
             "/opt/click.ubuntu.com/%s/%s/README" % (name, version)],
            universal_newlines=True)
        self.assertEqual(name, json.loads(output)["name"])

    def test_info_different_extension(self):
        name = "org.example.info"
        raw_path = self._make_click(name)
        path = "%s.extra" % raw_path
        os.rename(raw_path, path)
        output = subprocess.check_output([
            self.click_binary, "info", path], universal_newlines=True)
        self.assertEqual(name, json.loads(output)["name"])



# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""click commands."""

import importlib


all_commands = (
    "build",
    "buildsource",
    "chroot",
    "contents",
    "desktophook",
    "framework",
    "hook",
    "info",
    "install",
    "list",
    "pkgdir",
    "register",
    "unregister",
    "verify",
    )


hidden_commands = (
    "desktophook",
    )


def load_command(command):
    return importlib.import_module("click_package.commands.%s" % command)


def help_text():
    lines = []
    for command in all_commands:
        if command in hidden_commands:
            continue
        mod = load_command(command)
        lines.append("  %-21s %s" % (command, mod.__doc__.splitlines()[0]))
    return "\n".join(lines)



# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unregister an installed Click package for a user."""

from __future__ import print_function

from optparse import OptionParser
import os
import sys

from gi.repository import Click


def run(argv):
    parser = OptionParser("%prog unregister [options] PACKAGE-NAME [VERSION]")
    parser.add_option(
        "--root", metavar="PATH", help="look for additional packages in PATH")
    parser.add_option(
        "--user", metavar="USER",
        help="unregister package for USER (default: $SUDO_USER, if known)")
    parser.add_option(
        "--all-users", default=False, action="store_true",
        help="unregister package that was previously registered for all users")
    options, args = parser.parse_args(argv)
    if len(args) < 1:
        parser.error("need package name")
    if os.geteuid() != 0:
        parser.error(
            "click unregister must be started as root, since it may need to "
            "remove packages from disk")
    if options.user is None and "SUDO_USER" in os.environ:
        options.user = os.environ["SUDO_USER"]
    db = Click.DB()
    db.read(db_dir=None)
    if options.root is not None:
        db.add(options.root)
    package = args[0]
    if options.all_users:
        registry = Click.User.for_all_users(db)
    else:
        registry = Click.User.for_user(db, name=options.user)
    old_version = registry.get_version(package)
    if len(args) >= 2 and old_version != args[1]:
        print(
            "Not removing %s %s; expected version %s" %
            (package, old_version, args[1]),
            file=sys.stderr)
        sys.exit(1)
    registry.remove(package)
    db.maybe_remove(package, old_version)
    # TODO: remove data
    return 0



# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt <michael.vogt@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Integration tests for the click CLI frameworks command."""

import os
import subprocess

from .helpers import ClickTestCase


class TestFrameworks(ClickTestCase):
    def setUp(self):
        super(TestFrameworks, self).setUp()
        if (not os.path.exists("/usr/share/click/frameworks") or
                not os.listdir("/usr/share/click/frameworks")):
            self.skipTest("Please install ubuntu-sdk-libs")

    def test_framework_list(self):
        output = subprocess.check_output([
            self.click_binary, "framework", "list"], universal_newlines=True)
        self.assertTrue("ubuntu-sdk-" in output)



# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Preinst for Click packages.

In general there is a rule that Click packages may not have maintainer
scripts.  However, there is one exception: a static preinst used to cause
dpkg to fail if people attempt to install Click packages directly using dpkg
rather than via "click install".  This avoids accidents, since Click
packages use a different root of their filesystem tarball.
"""

from __future__ import print_function

__metaclass__ = type
__all__ = [
    'static_preinst',
    'static_preinst_matches',
    ]


_older_static_preinst = """\
#! /bin/sh
echo "Click packages may not be installed directly using dpkg."
echo "Use click-install instead."
exit 1
"""


_old_static_preinst = """\
#! /bin/sh
echo "Click packages may not be installed directly using dpkg."
echo "Use 'click-package install' instead."
exit 1
"""


static_preinst = """\
#! /bin/sh
echo "Click packages may not be installed directly using dpkg."
echo "Use 'click install' instead."
exit 1
"""


def static_preinst_matches(preinst):
    for allow_preinst in (
            _older_static_preinst,
            _old_static_preinst,
            static_preinst,
            ):
        if preinst == allow_preinst.encode():
            return True
    return False



# Copyright (C) 2014 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""A collection of variously hacky ctypes definitions for use with gimock."""

import ctypes

from click_package.tests.config import (
    STAT_OFFSET_GID,
    STAT_OFFSET_UID,
    STAT64_OFFSET_GID,
    STAT64_OFFSET_UID,
    )


class Passwd(ctypes.Structure):
    _fields_ = [
        ("pw_name", ctypes.c_char_p),
        ("pw_passwd", ctypes.c_char_p),
        ("pw_uid", ctypes.c_uint32),
        ("pw_gid", ctypes.c_uint32),
        ("pw_gecos", ctypes.c_char_p),
        ("pw_dir", ctypes.c_char_p),
        ("pw_shell", ctypes.c_char_p),
        ]


# TODO: This is pretty awful.  The layout of "struct stat" is complicated
# enough that we have to use offsetof() in configure to pick out the fields
# we care about.  Fortunately, we only care about a couple of fields, and
# since this is an output parameter it doesn't matter if our structure is
# too short (if we cared about this then we could use AC_CHECK_SIZEOF to
# figure it out).
class Stat(ctypes.Structure):
    _pack_ = 1
    _fields_ = []
    _fields_.append(
        ("pad0", ctypes.c_ubyte * min(STAT_OFFSET_UID, STAT_OFFSET_GID)))
    if STAT_OFFSET_UID < STAT_OFFSET_GID:
        _fields_.append(("st_uid", ctypes.c_uint32))
        pad = (STAT_OFFSET_GID - STAT_OFFSET_UID -
               ctypes.sizeof(ctypes.c_uint32))
        assert pad >= 0
        if pad > 0:
            _fields_.append(("pad1", ctypes.c_ubyte * pad))
        _fields_.append(("st_gid", ctypes.c_uint32))
    else:
        _fields_.append(("st_gid", ctypes.c_uint32))
        pad = (STAT_OFFSET_UID - STAT_OFFSET_GID -
               ctypes.sizeof(ctypes.c_uint32))
        assert pad >= 0
        if pad > 0:
            _fields_.append(("pad1", ctypes.c_ubyte * pad))
        _fields_.append(("st_uid", ctypes.c_uint32))


class Stat64(ctypes.Structure):
    _pack_ = 1
    _fields_ = []
    _fields_.append(
        ("pad0", ctypes.c_ubyte * min(STAT64_OFFSET_UID, STAT64_OFFSET_GID)))
    if STAT64_OFFSET_UID < STAT64_OFFSET_GID:
        _fields_.append(("st_uid", ctypes.c_uint32))
        pad = (STAT64_OFFSET_GID - STAT64_OFFSET_UID -
               ctypes.sizeof(ctypes.c_uint32))
        assert pad >= 0
        if pad > 0:
            _fields_.append(("pad1", ctypes.c_ubyte * pad))
        _fields_.append(("st_gid", ctypes.c_uint32))
    else:
        _fields_.append(("st_gid", ctypes.c_uint32))
        pad = (STAT64_OFFSET_UID - STAT64_OFFSET_GID -
               ctypes.sizeof(ctypes.c_uint32))
        assert pad >= 0
        if pad > 0:
            _fields_.append(("pad1", ctypes.c_ubyte * pad))
        _fields_.append(("st_uid", ctypes.c_uint32))



# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt <michael.vogt@canonical.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Pure python click framework handling support."""

import logging
import os
import re

try:
    import apt_pkg
except:
    pass

import click_package.paths


class ClickFrameworkInvalid(Exception):
    pass


# python version of the vala parse_deb822_file()
def parse_deb822_file(filename):
    data = {}
    with open(filename) as f:
        for line in f:
            line = line.strip()
            # from deb822.vala
            field_re_posix = (
                r'^([^:[:space:]]+)[[:space:]]*:[[:space:]]'
                '([^[:space:]].*?)[[:space:]]*$')
            # python does not do posix char classes
            field_re = field_re_posix.replace("[:space:]", "\s")
            blank_re_posix = r'^[[:space:]]*$'
            blank_re = blank_re_posix.replace("[:space:]", "\s")
            if re.match(blank_re, line):
                break
            match = re.match(field_re, line)
            if match and match.group(1) and match.group(2):
                data[match.group(1).lower()] = match.group(2)
    return data


# python version of vala get_frameworks_dir
def get_frameworks_dir():
    return click_package.paths.frameworks_dir


def get_framework_path(framework_name):
    framework_path = os.path.join(
        get_frameworks_dir(), framework_name+".framework")
    return framework_path


# python version of the vala click_framework_get_base_version()
def click_framework_get_base_version(framework_name):
    deb822 = parse_deb822_file(get_framework_path(framework_name))
    return deb822.get("base-version", None)


# python version of the vala click_framework_get_base_version()
def click_framework_get_base_name(framework_name):
    deb822 = parse_deb822_file(get_framework_path(framework_name))
    return deb822.get("base-name", None)


# python version of the vala click_framework_has_framework
def click_framework_has_framework(framework_name):
    return os.path.exists(get_framework_path(framework_name))


def validate_framework(framework_string, ignore_missing_frameworks=False):
    try:
        apt_pkg
    except NameError:
        logging.warning("No apt_pkg module, skipping validate_framework")
        return

    try:
        parsed_framework = apt_pkg.parse_depends(framework_string)
    except ValueError:
        raise ClickFrameworkInvalid(
            'Could not parse framework "%s"' % framework_string)

    base_name_versions = {}
    missing_frameworks = []
    for or_dep in parsed_framework:
        if len(or_dep) > 1:
            raise ClickFrameworkInvalid(
                'Alternative dependencies in framework "%s" not yet '
                'allowed' % framework_string)
        if or_dep[0][1] or or_dep[0][2]:
            raise ClickFrameworkInvalid(
                'Version relationship in framework "%s" not yet allowed' %
                framework_string)
        # now verify that different base versions are not mixed
        framework_name = or_dep[0][0]
        if not click_framework_has_framework(framework_name):
            missing_frameworks.append(framework_name)
            continue
        # ensure we do not use different base versions for the same base-name
        framework_base_name = click_framework_get_base_name(
            framework_name)
        framework_base_version = click_framework_get_base_version(
            framework_name)
        prev =  base_name_versions.get(framework_base_name, None)
        if prev and prev != framework_base_version:
            raise ClickFrameworkInvalid(
                'Multiple frameworks with different base versions are not '
                'allowed. Found: {} ({} != {})'.format(
                    framework_base_name,
                    framework_base_version,
                    base_name_versions[framework_base_name]))
        base_name_versions[framework_base_name] = framework_base_version

    if not ignore_missing_frameworks:
        if len(missing_frameworks) > 1:
            raise ClickFrameworkInvalid(
                'Frameworks %s not present on system (use '
                '--force-missing-framework option to override)' %
                ", ".join('"%s"' % f for f in missing_frameworks))
        elif missing_frameworks:
            raise ClickFrameworkInvalid(
                'Framework "%s" not present on system (use '
                '--force-missing-framework option to override)' %
                missing_frameworks[0])
    else:
        if len(missing_frameworks) > 1:
            logging.warning("Ignoring missing frameworks %s" % (
                ", ".join('"%s"' % f for f in missing_frameworks)))
        elif missing_frameworks:
            logging.warning('Ignoring missing framework "%s"' % (
                missing_frameworks[0]))



# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt <michael.vogt@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Integration tests for the click CLI build command."""

import os

from .helpers import ClickTestCase


class TestBuild(ClickTestCase):
    def test_build(self):
        path_to_click = self._make_click()
        self.assertTrue(os.path.exists(path_to_click))



# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt <michael.vogt@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Integration tests for the click CLI buildsource command."""

import os
import shutil
import subprocess
import tarfile
import tempfile

from .helpers import (
    chdir,
    ClickTestCase,
)


class TestBuildSource(ClickTestCase):

    def test_buildsource(self):
        temp_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, temp_dir)
        with chdir(temp_dir):
            with open(os.path.join(temp_dir, "README"), "w") as f:
                f.write("I'm a source package")
            os.mkdir(os.path.join(temp_dir, ".git"))
            os.mkdir(os.path.join(temp_dir, ".bzr"))
            os.mkdir(os.path.join(temp_dir, ".normal"))
            self._create_manifest(os.path.join(temp_dir, "manifest.json"),
                                  "srcfoo", "1.2", "ubuntu-sdk-13.10")
            subprocess.check_call(
                [self.click_binary, "buildsource", temp_dir],
                universal_newlines=True)
            # ensure we have the content we expect
            source_file = "srcfoo_1.2.tar.gz"
            tar = tarfile.open(source_file)
            self.assertEqual(
                sorted(tar.getnames()),
                sorted([".", "./.normal", "./manifest.json", "./README"]))



#! /usr/bin/python3

# Copyright (C) 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Print the directory where a Click package is unpacked."""

from __future__ import print_function

from optparse import OptionParser
import sys

from gi.repository import Click


def run(argv):
    parser = OptionParser("%prog pkgdir [options] {PACKAGE-NAME|PATH}")
    parser.add_option(
        "--root", metavar="PATH", help="look for additional packages in PATH")
    parser.add_option(
        "--user", metavar="USER",
        help="look up PACKAGE-NAME for USER (if you have permission; "
             "default: current user)")
    options, args = parser.parse_args(argv)
    if len(args) < 1:
        parser.error("need package name")
    try:
        if "/" in args[0]:
            print(Click.find_package_directory(args[0]))
        else:
            db = Click.DB()
            db.read(db_dir=None)
            if options.root is not None:
                db.add(options.root)
            package_name = args[0]
            registry = Click.User.for_user(db, name=options.user)
            print(registry.get_path(package_name))
    except Exception as e:
        print(e, file=sys.stderr)
        return 1
    return 0



# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt <michael.vogt@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Integration tests for the click CLI contents command."""

import re
import subprocess

from .helpers import ClickTestCase


class TestContents(ClickTestCase):
    def test_contents(self):
        name = "com.example.contents"
        path_to_click = self._make_click(name)
        output = subprocess.check_output([
            self.click_binary, "contents", path_to_click],
            universal_newlines=True)
        self.assertTrue(re.search(
            r'-rw-r[-w]-r-- root/root\s+[0-9]+\s+[0-9-]+ [0-9:]+ ./README',
            output))



# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for click_package.hooks."""

from __future__ import print_function

__metaclass__ = type
__all__ = [
    "TestClickHookSystemLevel",
    "TestClickHookUserLevel",
    "TestClickPatternFormatter",
    "TestPackageInstallHooks",
    "TestPackageRemoveHooks",
    ]


from functools import partial
from itertools import takewhile
import json
import os
from textwrap import dedent

from gi.repository import Click, GLib

from click_package.tests.gimock_types import Passwd
from click_package.tests.helpers import TestCase, mkfile, mkfile_utf8


class TestClickPatternFormatter(TestCase):
    def _make_variant(self, **kwargs):
        # pygobject's Variant creator can't handle maybe types, so we have
        # to do this by hand.
        builder = GLib.VariantBuilder.new(GLib.VariantType.new("a{sms}"))
        for key, value in kwargs.items():
            entry = GLib.VariantBuilder.new(GLib.VariantType.new("{sms}"))
            entry.add_value(GLib.Variant.new_string(key))
            entry.add_value(GLib.Variant.new_maybe(
                GLib.VariantType.new("s"),
                None if value is None else GLib.Variant.new_string(value)))
            builder.add_value(entry.end())
        return builder.end()

    def test_expands_provided_keys(self):
        self.assertEqual(
            "foo.bar",
            Click.pattern_format("foo.${key}", self._make_variant(key="bar")))
        self.assertEqual(
            "foo.barbaz",
            Click.pattern_format(
                "foo.${key1}${key2}",
                self._make_variant(key1="bar", key2="baz")))

    def test_expands_missing_keys_to_empty_string(self):
        self.assertEqual(
            "xy", Click.pattern_format("x${key}y", self._make_variant()))

    def test_preserves_unmatched_dollar(self):
        self.assertEqual("$", Click.pattern_format("$", self._make_variant()))
        self.assertEqual(
            "$ {foo}", Click.pattern_format("$ {foo}", self._make_variant()))
        self.assertEqual(
            "x${y",
            Click.pattern_format("${key}${y", self._make_variant(key="x")))

    def test_double_dollar(self):
        self.assertEqual("$", Click.pattern_format("$$", self._make_variant()))
        self.assertEqual(
            "${foo}", Click.pattern_format("$${foo}", self._make_variant()))
        self.assertEqual(
            "x$y",
            Click.pattern_format("x$$${key}", self._make_variant(key="y")))

    def test_possible_expansion(self):
        self.assertEqual(
            {"id": "abc"},
            Click.pattern_possible_expansion(
                "x_abc_1", "x_${id}_${num}",
                self._make_variant(num="1")).unpack())
        self.assertIsNone(
            Click.pattern_possible_expansion(
                "x_abc_1", "x_${id}_${num}", self._make_variant(num="2")))


class TestClickHookBase(TestCase):

    TEST_USER = "test-user"

    def setUp(self):
        super(TestClickHookBase, self).setUp()
        self.use_temp_dir()
        self.db = Click.DB()
        self.db.add(self.temp_dir)
        self.spawn_calls = []

    def _make_installed_click(self, package="test-1", version="1.0",
                              json_data={},
                              make_current=True,
                              all_users=False):
        with mkfile_utf8(os.path.join(
                self.temp_dir, package, version, ".click", "info",
                "%s.manifest" % package)) as f:
            json.dump(json_data, f, ensure_ascii=False)
        if make_current:
            os.symlink(
                version, os.path.join(self.temp_dir, package, "current"))
        if all_users:
            db = Click.User.for_all_users(self.db)
        else:
            db = Click.User.for_user(self.db, self.TEST_USER)
        db.set_version(package, version)

    def _make_hook_file(self, content, hookname="test"):
        hook_file = os.path.join(self.hooks_dir, "%s.hook" % hookname)
        with mkfile(hook_file) as f:
            print(content, file=f)

    def _setup_hooks_dir(self, preloads, hooks_dir=None):
        if hooks_dir is None:
            hooks_dir = self.temp_dir
        preloads["click_get_hooks_dir"].side_effect = (
            lambda: self.make_string(hooks_dir))
        self.hooks_dir = hooks_dir

    def g_spawn_sync_side_effect(self, status_map, working_directory, argv,
                                 envp, flags, child_setup, user_data,
                                 standard_output, standard_error, exit_status,
                                 error):
        self.spawn_calls.append(list(takewhile(lambda x: x is not None, argv)))
        if argv[0] in status_map:
            exit_status[0] = status_map[argv[0]]
        else:
            self.delegate_to_original("g_spawn_sync")
        return 0


class TestClickHookSystemLevel(TestClickHookBase):
    def test_open(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            self._make_hook_file(dedent("""\
                    Pattern: /usr/share/test/${id}.test
                    # Comment
                    Exec: test-update
                    User: root
                    """))
            hook = Click.Hook.open(self.db, "test")
            self.assertCountEqual(
                ["pattern", "exec", "user"], hook.get_fields())
            self.assertEqual(
                "/usr/share/test/${id}.test", hook.get_field("pattern"))
            self.assertEqual("test-update", hook.get_field("exec"))
            self.assertRaisesHooksError(
                Click.HooksError.MISSING_FIELD, hook.get_field, "nonexistent")
            self.assertFalse(hook.props.is_user_level)

    def test_open_unopenable_file(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            os.symlink("nonexistent", os.path.join(self.hooks_dir, "foo.hook"))
            self.assertRaisesHooksError(
                Click.HooksError.NO_SUCH_HOOK, Click.Hook.open, self.db, "foo")

    def test_hook_name_absent(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            self._make_hook_file(
                "Pattern: /usr/share/test/${id}.test")
            hook = Click.Hook.open(self.db, "test")
            self.assertEqual("test", hook.get_hook_name())

    def test_hook_name_present(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            self._make_hook_file(dedent("""\
                Pattern: /usr/share/test/${id}.test
                Hook-Name: other"""))
            hook = Click.Hook.open(self.db, "test")
            self.assertEqual("other", hook.get_hook_name())

    def test_invalid_app_id(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            self._make_hook_file(dedent("""\
                    Pattern: /usr/share/test/${id}.test
                    # Comment
                    Exec: test-update
                    User: root
            """))
            hook = Click.Hook.open(self.db, "test")
            self.assertRaisesHooksError(
                Click.HooksError.BAD_APP_NAME, hook.get_app_id,
                "package", "0.1", "app_name")
            self.assertRaisesHooksError(
                Click.HooksError.BAD_APP_NAME, hook.get_app_id,
                "package", "0.1", "app/name")
            self.assertRaisesHooksError(
                Click.HooksError.BAD_APP_NAME, hook.get_pattern,
                "package", "0.1", "app_name")
            self.assertRaisesHooksError(
                Click.HooksError.BAD_APP_NAME, hook.get_pattern,
                "package", "0.1", "app/name")

    def test_short_id_invalid(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            self._make_hook_file(
                "Pattern: /usr/share/test/${short-id}.test")
            hook = Click.Hook.open(self.db, "test")
            # It would perhaps be better if unrecognised $-expansions raised
            # KeyError, but they don't right now.
            self.assertEqual(
                "/usr/share/test/.test",
                hook.get_pattern("package", "0.1", "app-name", user_name=None))

    def test_short_id_valid_with_single_version(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            self._make_hook_file(dedent("""\
                Pattern: /usr/share/test/${short-id}.test
                Single-Version: yes"""))
            hook = Click.Hook.open(self.db, "test")
            self.assertEqual(
                "/usr/share/test/package_app-name.test",
                hook.get_pattern("package", "0.1", "app-name", user_name=None))

    def test_run_commands(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir", "g_spawn_sync") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            preloads["g_spawn_sync"].side_effect = partial(
                self.g_spawn_sync_side_effect, {b"/bin/sh": 0})
            with mkfile(os.path.join(self.temp_dir, "test.hook")) as f:
                print("Exec: test-update", file=f)
                print("User: root", file=f)
            hook = Click.Hook.open(self.db, "test")
            self.assertEqual(
                "root", hook.get_run_commands_user(user_name=None))
            hook.run_commands(user_name=None)
            self.assertEqual(
                [[b"/bin/sh", b"-c", b"test-update"]], self.spawn_calls)

    def test_run_commands_fail(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir", "g_spawn_sync") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            preloads["g_spawn_sync"].side_effect = partial(
                self.g_spawn_sync_side_effect, {b"/bin/sh": 1})
            with mkfile(os.path.join(self.temp_dir, "test.hook")) as f:
                print("Exec: test-update", file=f)
                print("User: root", file=f)
            hook = Click.Hook.open(self.db, "test")
            self.assertRaisesHooksError(
                Click.HooksError.COMMAND_FAILED, hook.run_commands,
                user_name=None)

    def test_install_package(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            self._make_hook_file(
                "Pattern: %s/${id}.test" % self.temp_dir)
            os.makedirs(
                os.path.join(self.temp_dir, "org.example.package", "1.0"))
            hook = Click.Hook.open(self.db, "test")
            hook.install_package(
                "org.example.package", "1.0", "test-app", "foo/bar",
                user_name=None)
            symlink_path = os.path.join(
                self.temp_dir, "org.example.package_test-app_1.0.test")
            target_path = os.path.join(
                self.temp_dir, "org.example.package", "1.0", "foo", "bar")
            self.assertTrue(os.path.islink(symlink_path))
            self.assertEqual(target_path, os.readlink(symlink_path))

    def test_install_package_trailing_slash(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            self._make_hook_file(
                "Pattern: %s/${id}/" % self.temp_dir)
            os.makedirs(
                os.path.join(self.temp_dir, "org.example.package", "1.0"))
            hook = Click.Hook.open(self.db, "test")
            hook.install_package(
                "org.example.package", "1.0", "test-app", "foo",
                user_name=None)
            symlink_path = os.path.join(
                self.temp_dir, "org.example.package_test-app_1.0")
            target_path = os.path.join(
                self.temp_dir, "org.example.package", "1.0", "foo")
            self.assertTrue(os.path.islink(symlink_path))
            self.assertEqual(target_path, os.readlink(symlink_path))

    def test_install_package_uses_deepest_copy(self):
        # If the same version of a package is unpacked in multiple
        # databases, then we make sure the link points to the deepest copy,
        # even if it already points somewhere else.  It is important to be
        # consistent about this since system hooks may only have a single
        # target for any given application ID.
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            self._make_hook_file(
                "Pattern: %s/${id}.test" % self.temp_dir)
            underlay = os.path.join(self.temp_dir, "underlay")
            overlay = os.path.join(self.temp_dir, "overlay")
            db = Click.DB()
            db.add(underlay)
            db.add(overlay)
            os.makedirs(os.path.join(underlay, "org.example.package", "1.0"))
            os.makedirs(os.path.join(overlay, "org.example.package", "1.0"))
            symlink_path = os.path.join(
                self.temp_dir, "org.example.package_test-app_1.0.test")
            underlay_target_path = os.path.join(
                underlay, "org.example.package", "1.0", "foo")
            overlay_target_path = os.path.join(
                overlay, "org.example.package", "1.0", "foo")
            os.symlink(overlay_target_path, symlink_path)
            hook = Click.Hook.open(db, "test")
            hook.install_package(
                "org.example.package", "1.0", "test-app", "foo",
                user_name=None)
            self.assertTrue(os.path.islink(symlink_path))
            self.assertEqual(underlay_target_path, os.readlink(symlink_path))

    def test_upgrade(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            self._make_hook_file(
                "Pattern: %s/${id}.test" % self.temp_dir)
            symlink_path = os.path.join(
                self.temp_dir, "org.example.package_test-app_1.0.test")
            os.symlink("old-target", symlink_path)
            os.makedirs(
                os.path.join(self.temp_dir, "org.example.package", "1.0"))
            hook = Click.Hook.open(self.db, "test")
            hook.install_package(
                "org.example.package", "1.0", "test-app", "foo/bar",
                user_name=None)
            target_path = os.path.join(
                self.temp_dir, "org.example.package", "1.0", "foo", "bar")
            self.assertTrue(os.path.islink(symlink_path))
            self.assertEqual(target_path, os.readlink(symlink_path))

    def test_remove_package(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            self._make_hook_file(
                "Pattern: %s/${id}.test" % self.temp_dir)
            symlink_path = os.path.join(
                self.temp_dir, "org.example.package_test-app_1.0.test")
            os.symlink("old-target", symlink_path)
            hook = Click.Hook.open(self.db, "test")
            hook.remove_package(
                "org.example.package", "1.0", "test-app", user_name=None)
            self.assertFalse(os.path.exists(symlink_path))

    def test_install(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(
                preloads, hooks_dir=os.path.join(self.temp_dir, "hooks"))
            self._make_hook_file(
                "Pattern: %s/${id}.new" % self.temp_dir,
                hookname="new")
            self._make_installed_click("test-1", "1.0", json_data={
                    "maintainer":
                        b"Unic\xc3\xb3de <unicode@example.org>".decode(
                            "UTF-8"),
                    "hooks": {"test1-app": {"new": "target-1"}}})
            self._make_installed_click("test-2", "2.0", json_data={
                    "maintainer":
                        b"Unic\xc3\xb3de <unicode@example.org>".decode(
                            "UTF-8"),
                    "hooks": {"test1-app": {"new": "target-2"}},
                })
            hook = Click.Hook.open(self.db, "new")
            hook.install(user_name=None)
            path_1 = os.path.join(self.temp_dir, "test-1_test1-app_1.0.new")
            self.assertTrue(os.path.lexists(path_1))
            self.assertEqual(
                os.path.join(self.temp_dir, "test-1", "1.0", "target-1"),
                os.readlink(path_1))
            path_2 = os.path.join(self.temp_dir, "test-2_test1-app_2.0.new")
            self.assertTrue(os.path.lexists(path_2))
            self.assertEqual(
                os.path.join(self.temp_dir, "test-2", "2.0", "target-2"),
                os.readlink(path_2))

    def test_remove(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(
                preloads, hooks_dir=os.path.join(self.temp_dir, "hooks"))
            self._make_hook_file(
                "Pattern: %s/${id}.old" % self.temp_dir,
                hookname="old")
            self._make_installed_click("test-1", "1.0", json_data={
                "hooks": {"test1-app": {"old": "target-1"}}})
            path_1 = os.path.join(self.temp_dir, "test-1_test1-app_1.0.old")
            os.symlink(
                os.path.join(self.temp_dir, "test-1", "1.0", "target-1"),
                path_1)
            self._make_installed_click("test-2", "2.0", json_data={
                "hooks": {"test2-app": {"old": "target-2"}}})
            path_2 = os.path.join(self.temp_dir, "test-2_test2-app_2.0.old")
            os.symlink(
                os.path.join(self.temp_dir, "test-2", "2.0", "target-2"),
                path_2)
            hook = Click.Hook.open(self.db, "old")
            hook.remove(user_name=None)
            self.assertFalse(os.path.exists(path_1))
            self.assertFalse(os.path.exists(path_2))

    def test_sync(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(
                preloads, hooks_dir=os.path.join(self.temp_dir, "hooks"))
            self._make_hook_file(
                "Pattern: %s/${id}.test" % self.temp_dir)
            self._make_installed_click("test-1", "1.0", json_data={
                "hooks": {"test1-app": {"test": "target-1"}}})
            self._make_installed_click(
                "test-2", "1.0", make_current=False,
                json_data={"hooks": {"test2-app": {"test": "target-2"}}})
            self._make_installed_click("test-2", "1.1", json_data={
                "hooks": {"test2-app": {"test": "target-2"}}})
            path_1 = os.path.join(self.temp_dir, "test-1_test1-app_1.0.test")
            os.symlink(
                os.path.join(self.temp_dir, "test-1", "1.0", "target-1"),
                path_1)
            path_2_1_0 = os.path.join(
                self.temp_dir, "test-2_test2-app_1.0.test")
            path_2_1_1 = os.path.join(
                self.temp_dir, "test-2_test2-app_1.1.test")
            path_3 = os.path.join(self.temp_dir, "test-3_test3-app_1.0.test")
            os.symlink(
                os.path.join(self.temp_dir, "test-3", "1.0", "target-3"),
                path_3)
            hook = Click.Hook.open(self.db, "test")
            hook.sync(user_name=None)
            self.assertTrue(os.path.lexists(path_1))
            self.assertEqual(
                os.path.join(self.temp_dir, "test-1", "1.0", "target-1"),
                os.readlink(path_1))
            self.assertTrue(os.path.lexists(path_2_1_0))
            self.assertEqual(
                os.path.join(self.temp_dir, "test-2", "1.0", "target-2"),
                os.readlink(path_2_1_0))
            self.assertTrue(os.path.lexists(path_2_1_1))
            self.assertEqual(
                os.path.join(self.temp_dir, "test-2", "1.1", "target-2"),
                os.readlink(path_2_1_1))
            self.assertFalse(os.path.lexists(path_3))


class TestClickHookUserLevel(TestClickHookBase):
    def test_open(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            self._make_hook_file(dedent("""\
                    User-Level: yes
                    Pattern: ${home}/.local/share/test/${id}.test
                    # Comment
                    Exec: test-update
                    """))
            hook = Click.Hook.open(self.db, "test")
            self.assertCountEqual(
                ["user-level", "pattern", "exec"], hook.get_fields())
            self.assertEqual(
                "${home}/.local/share/test/${id}.test",
                hook.get_field("pattern"))
            self.assertEqual("test-update", hook.get_field("exec"))
            self.assertRaisesHooksError(
                Click.HooksError.MISSING_FIELD, hook.get_field, "nonexistent")
            self.assertTrue(hook.props.is_user_level)

    def test_hook_name_absent(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            self._make_hook_file(dedent("""\
                User-Level: yes
                Pattern: ${home}/.local/share/test/${id}.test"""))
            hook = Click.Hook.open(self.db, "test")
            self.assertEqual("test", hook.get_hook_name())

    def test_hook_name_present(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            self._make_hook_file(dedent("""\
                User-Level: yes
                Pattern: ${home}/.local/share/test/${id}.test
                Hook-Name: other"""))
            hook = Click.Hook.open(self.db, "test")
            self.assertEqual("other", hook.get_hook_name())

    def test_invalid_app_id(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            self._make_hook_file(dedent("""\
                    User-Level: yes
                    Pattern: ${home}/.local/share/test/${id}.test
                    # Comment
                    Exec: test-update"""))
            hook = Click.Hook.open(self.db, "test")
            self.assertRaisesHooksError(
                Click.HooksError.BAD_APP_NAME, hook.get_app_id,
                "package", "0.1", "app_name")
            self.assertRaisesHooksError(
                Click.HooksError.BAD_APP_NAME, hook.get_app_id,
                "package", "0.1", "app/name")
            self.assertRaisesHooksError(
                Click.HooksError.BAD_APP_NAME, hook.get_pattern,
                "package", "0.1", "app_name")
            self.assertRaisesHooksError(
                Click.HooksError.BAD_APP_NAME, hook.get_pattern,
                "package", "0.1", "app/name")

    def test_short_id_valid(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir", "getpwnam") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            preloads["getpwnam"].side_effect = (
                lambda name: self.make_pointer(Passwd(pw_dir=b"/mock")))
            self._make_hook_file(dedent("""\
                User-Level: yes
                Pattern: ${home}/.local/share/test/${short-id}.test
            """))
            hook = Click.Hook.open(self.db, "test")
            self.assertEqual(
                "/mock/.local/share/test/package_app-name.test",
                hook.get_pattern(
                    "package", "0.1", "app-name", user_name="mock"))

    def test_run_commands(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir", "g_spawn_sync") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            preloads["g_spawn_sync"].side_effect = partial(
                self.g_spawn_sync_side_effect, {b"/bin/sh": 0})
            with mkfile(os.path.join(self.temp_dir, "test.hook")) as f:
                print("User-Level: yes", file=f)
                print("Exec: test-update", file=f)
            hook = Click.Hook.open(self.db, "test")
            self.assertEqual(
                self.TEST_USER,
                hook.get_run_commands_user(user_name=self.TEST_USER))
            hook.run_commands(user_name=self.TEST_USER)
            self.assertEqual(
                [[b"/bin/sh", b"-c", b"test-update"]], self.spawn_calls)

    def test_run_commands_fail(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir", "g_spawn_sync") as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            preloads["g_spawn_sync"].side_effect = partial(
                self.g_spawn_sync_side_effect, {b"/bin/sh": 1})
            with mkfile(os.path.join(self.temp_dir, "test.hook")) as f:
                print("User-Level: yes", file=f)
                print("Exec: test-update", file=f)
            hook = Click.Hook.open(self.db, "test")
            self.assertRaisesHooksError(
                Click.HooksError.COMMAND_FAILED, hook.run_commands,
                user_name=self.TEST_USER)

    def test_install_package(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir", "click_get_user_home",
                ) as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            preloads["click_get_user_home"].return_value = b"/home/test-user"
            os.makedirs(os.path.join(
                self.temp_dir, "org.example.package", "1.0"))
            user_db = Click.User.for_user(self.db, self.TEST_USER)
            user_db.set_version("org.example.package", "1.0")
            self._make_hook_file(dedent("""\
                User-Level: yes
                Pattern: %s/${id}.test""") % self.temp_dir)
            hook = Click.Hook.open(self.db, "test")
            hook.install_package(
                "org.example.package", "1.0", "test-app", "foo/bar",
                user_name=self.TEST_USER)
            symlink_path = os.path.join(
                self.temp_dir, "org.example.package_test-app_1.0.test")
            target_path = os.path.join(
                self.temp_dir, ".click", "users", self.TEST_USER,
                "org.example.package", "foo", "bar")
            self.assertTrue(os.path.islink(symlink_path))
            self.assertEqual(target_path, os.readlink(symlink_path))

    def test_install_package_trailing_slash(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir", "click_get_user_home",
                ) as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            preloads["click_get_user_home"].return_value = b"/home/test-user"
            os.makedirs(os.path.join(
                self.temp_dir, "org.example.package", "1.0"))
            user_db = Click.User.for_user(self.db, self.TEST_USER)
            user_db.set_version("org.example.package", "1.0")
            self._make_hook_file(dedent("""\
                User-Level: yes
                Pattern: %s/${id}/""") % self.temp_dir)
            hook = Click.Hook.open(self.db, "test")
            hook.install_package(
                "org.example.package", "1.0", "test-app", "foo",
                user_name=self.TEST_USER)
            symlink_path = os.path.join(
                self.temp_dir, "org.example.package_test-app_1.0")
            target_path = os.path.join(
                self.temp_dir, ".click", "users", self.TEST_USER,
                "org.example.package", "foo")
            self.assertTrue(os.path.islink(symlink_path))
            self.assertEqual(target_path, os.readlink(symlink_path))

    def test_install_package_removes_previous(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir", "click_get_user_home",
                ) as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            preloads["click_get_user_home"].return_value = b"/home/test-user"
            os.makedirs(os.path.join(
                self.temp_dir, "org.example.package", "1.0"))
            os.makedirs(os.path.join(
                self.temp_dir, "org.example.package", "1.1"))
            user_db = Click.User.for_user(self.db, self.TEST_USER)
            user_db.set_version("org.example.package", "1.0")
            self._make_hook_file(dedent("""\
                User-Level: yes
                Pattern: %s/${id}.test""") % self.temp_dir)
            hook = Click.Hook.open(self.db, "test")
            hook.install_package(
                "org.example.package", "1.0", "test-app", "foo/bar",
                user_name=self.TEST_USER)
            hook.install_package(
                "org.example.package", "1.1", "test-app", "foo/bar",
                user_name=self.TEST_USER)
            old_symlink_path = os.path.join(
                self.temp_dir, "org.example.package_test-app_1.0.test")
            symlink_path = os.path.join(
                self.temp_dir, "org.example.package_test-app_1.1.test")
            self.assertFalse(os.path.islink(old_symlink_path))
            self.assertTrue(os.path.islink(symlink_path))
            target_path = os.path.join(
                self.temp_dir, ".click", "users", self.TEST_USER,
                "org.example.package", "foo", "bar")
            self.assertEqual(target_path, os.readlink(symlink_path))

    def test_upgrade(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir", "click_get_user_home",
                ) as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            preloads["click_get_user_home"].return_value = b"/home/test-user"
            symlink_path = os.path.join(
                self.temp_dir, "org.example.package_test-app_1.0.test")
            os.symlink("old-target", symlink_path)
            os.makedirs(os.path.join(
                self.temp_dir, "org.example.package", "1.0"))
            user_db = Click.User.for_user(self.db, self.TEST_USER)
            user_db.set_version("org.example.package", "1.0")
            self._make_hook_file(dedent("""\
                User-Level: yes
                Pattern: %s/${id}.test""") % self.temp_dir)
            hook = Click.Hook.open(self.db, "test")
            hook.install_package(
                "org.example.package", "1.0", "test-app", "foo/bar",
                user_name=self.TEST_USER)
            target_path = os.path.join(
                self.temp_dir, ".click", "users", self.TEST_USER,
                "org.example.package", "foo", "bar")
            self.assertTrue(os.path.islink(symlink_path))
            self.assertEqual(target_path, os.readlink(symlink_path))

    def test_remove_package(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir", "click_get_user_home",
                ) as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            preloads["click_get_user_home"].return_value = b"/home/test-user"
            self._make_hook_file(dedent("""\
                User-Level: yes
                Pattern: %s/${id}.test""") % self.temp_dir)
            symlink_path = os.path.join(
                self.temp_dir, "org.example.package_test-app_1.0.test")
            os.symlink("old-target", symlink_path)
            hook = Click.Hook.open(self.db, "test")
            hook.remove_package(
                "org.example.package", "1.0", "test-app",
                user_name=self.TEST_USER)
            self.assertFalse(os.path.exists(symlink_path))

    def test_install(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir", "click_get_user_home", "getpwnam"
                ) as (enter, preloads):
            enter()
            # Don't tell click about the hooks directory yet.
            self._setup_hooks_dir(preloads)
            preloads["click_get_user_home"].return_value = b"/home/test-user"
            preloads["getpwnam"].side_effect = (
                lambda name: self.make_pointer(Passwd(pw_uid=1, pw_gid=1)))
            with mkfile(os.path.join(self.temp_dir, "hooks", "new.hook")) as f:
                print("User-Level: yes", file=f)
                print("Pattern: %s/${id}.new" % self.temp_dir, file=f)
            self._make_installed_click("test-1", "1.0", json_data={
                    "maintainer":
                        b"Unic\xc3\xb3de <unicode@example.org>".decode(
                            "UTF-8"),
                    "hooks": {"test1-app": {"new": "target-1"}},
            })
            self._make_installed_click("test-2", "2.0", json_data={
                    "maintainer":
                        b"Unic\xc3\xb3de <unicode@example.org>".decode(
                            "UTF-8"),
                    "hooks": {"test1-app": {"new": "target-2"}},
                })
            # Now tell click about the hooks directory and make sure it
            # catches up correctly.
            self._setup_hooks_dir(
                preloads, hooks_dir=os.path.join(self.temp_dir, "hooks"))
            hook = Click.Hook.open(self.db, "new")
            hook.install(user_name=None)
            path_1 = os.path.join(self.temp_dir, "test-1_test1-app_1.0.new")
            self.assertTrue(os.path.lexists(path_1))
            self.assertEqual(
                os.path.join(
                    self.temp_dir, ".click", "users", self.TEST_USER, "test-1",
                    "target-1"),
                os.readlink(path_1))
            path_2 = os.path.join(self.temp_dir, "test-2_test1-app_2.0.new")
            self.assertTrue(os.path.lexists(path_2))
            self.assertEqual(
                os.path.join(
                    self.temp_dir, ".click", "users", self.TEST_USER, "test-2",
                    "target-2"),
                os.readlink(path_2))

            os.unlink(path_1)
            os.unlink(path_2)
            hook.install(user_name="another-user")
            self.assertFalse(os.path.lexists(path_1))
            self.assertFalse(os.path.lexists(path_2))

            hook.install(user_name=self.TEST_USER)
            self.assertTrue(os.path.lexists(path_1))
            self.assertEqual(
                os.path.join(
                    self.temp_dir, ".click", "users", self.TEST_USER, "test-1",
                    "target-1"),
                os.readlink(path_1))
            self.assertTrue(os.path.lexists(path_2))
            self.assertEqual(
                os.path.join(
                    self.temp_dir, ".click", "users", self.TEST_USER, "test-2",
                    "target-2"),
                os.readlink(path_2))

    def test_remove(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir", "click_get_user_home",
                ) as (enter, preloads):
            enter()
            # Don't tell click about the hooks directory yet.
            self._setup_hooks_dir(preloads)
            preloads["click_get_user_home"].return_value = b"/home/test-user"
            with mkfile(os.path.join(self.temp_dir, "hooks", "old.hook")) as f:
                print("User-Level: yes", file=f)
                print("Pattern: %s/${id}.old" % self.temp_dir, file=f)
            user_db = Click.User.for_user(self.db, self.TEST_USER)
            self._make_installed_click("test-1", "1.0", json_data={
                "hooks": {"test1-app": {"old": "target-1"}}})
            path_1 = os.path.join(self.temp_dir, "test-1_test1-app_1.0.old")
            os.symlink(
                os.path.join(user_db.get_path("test-1"), "target-1"), path_1)
            self._make_installed_click("test-2", "2.0", json_data={
                "hooks": {"test2-app": {"old": "target-2"}}})
            path_2 = os.path.join(self.temp_dir, "test-2_test2-app_2.0.old")
            os.symlink(
                os.path.join(user_db.get_path("test-2"), "target-2"), path_2)
            # Now tell click about the hooks directory and make sure it
            # catches up correctly.
            self._setup_hooks_dir(
                preloads, hooks_dir=os.path.join(self.temp_dir, "hooks"))
            hook = Click.Hook.open(self.db, "old")
            hook.remove(user_name=None)
            self.assertFalse(os.path.exists(path_1))
            self.assertFalse(os.path.exists(path_2))

    def test_sync(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir", "click_get_user_home",
                ) as (enter, preloads):
            enter()
            preloads["click_get_user_home"].return_value = b"/home/test-user"
            self._setup_hooks_dir(preloads)
            with mkfile(
                    os.path.join(self.temp_dir, "hooks", "test.hook")) as f:
                print("User-Level: yes", file=f)
                print("Pattern: %s/${id}.test" % self.temp_dir, file=f)
            self._make_installed_click("test-1", "1.0", json_data={
                "hooks": {"test1-app": {"test": "target-1"}}})
            self._make_installed_click("test-2", "1.1", json_data={
                "hooks": {"test2-app": {"test": "target-2"}}})
            path_1 = os.path.join(self.temp_dir, "test-1_test1-app_1.0.test")
            os.symlink(
                os.path.join(
                    self.temp_dir, ".click", "users", self.TEST_USER, "test-1",
                    "target-1"),
                path_1)
            path_2 = os.path.join(self.temp_dir, "test-2_test2-app_1.1.test")
            path_3 = os.path.join(self.temp_dir, "test-3_test3-app_1.0.test")
            os.symlink(
                os.path.join(
                    self.temp_dir, ".click", "users", self.TEST_USER, "test-3",
                    "target-3"),
                path_3)
            self._setup_hooks_dir(
                preloads, hooks_dir=os.path.join(self.temp_dir, "hooks"))
            hook = Click.Hook.open(self.db, "test")
            hook.sync(user_name=self.TEST_USER)
            self.assertTrue(os.path.lexists(path_1))
            self.assertEqual(
                os.path.join(
                    self.temp_dir, ".click", "users", self.TEST_USER, "test-1",
                    "target-1"),
                os.readlink(path_1))
            self.assertTrue(os.path.lexists(path_2))
            self.assertEqual(
                os.path.join(
                    self.temp_dir, ".click", "users", self.TEST_USER, "test-2",
                    "target-2"),
                os.readlink(path_2))
            self.assertFalse(os.path.lexists(path_3))

    def test_sync_without_user_db(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir", "click_get_user_home",
                ) as (enter, preloads):
            enter()
            preloads["click_get_user_home"].return_value = b"/home/test-user"
            self._setup_hooks_dir(preloads)
            with mkfile(
                    os.path.join(self.temp_dir, "hooks", "test.hook")) as f:
                print("User-Level: yes", file=f)
                print("Pattern: %s/${id}.test" % self.temp_dir, file=f)
            self._make_installed_click(
                "test-package", "1.0", all_users=True, json_data={
                    "hooks": {"test-app": {"test": "target"}}})
            self._setup_hooks_dir(
                preloads, hooks_dir=os.path.join(self.temp_dir, "hooks"))
            hook = Click.Hook.open(self.db, "test")
            hook.sync(user_name=self.TEST_USER)
            self.assertFalse(os.path.exists(os.path.join(
                self.temp_dir, ".click", "users", self.TEST_USER,
                "test-package")))

    def test_sync_uses_deepest_copy(self):
        # If the same version of a package is unpacked in multiple
        # databases, then we make sure the user link points to the deepest
        # copy, even if it already points somewhere else.  It is important
        # to be consistent about this since system hooks may only have a
        # single target for any given application ID, and user links must
        # match system hooks so that (for example) the version of an
        # application run by a user has a matching system AppArmor profile.
        with self.run_in_subprocess(
                "click_get_hooks_dir", "click_get_user_home",
                ) as (enter, preloads):
            enter()
            self._setup_hooks_dir(preloads)
            preloads["click_get_user_home"].return_value = b"/home/test-user"
            with mkfile(os.path.join(self.temp_dir, "test.hook")) as f:
                print("User-Level: yes", file=f)
                print("Pattern: %s/${id}.test" % self.temp_dir, file=f)
            underlay = os.path.join(self.temp_dir, "underlay")
            overlay = os.path.join(self.temp_dir, "overlay")
            db = Click.DB()
            db.add(underlay)
            db.add(overlay)
            underlay_unpacked = os.path.join(underlay, "test-package", "1.0")
            overlay_unpacked = os.path.join(overlay, "test-package", "1.0")
            os.makedirs(underlay_unpacked)
            os.makedirs(overlay_unpacked)
            manifest = {"hooks": {"test-app": {"test": "foo"}}}
            with mkfile(os.path.join(
                    underlay_unpacked, ".click", "info",
                    "test-package.manifest")) as f:
                json.dump(manifest, f)
            with mkfile(os.path.join(
                    overlay_unpacked, ".click", "info",
                    "test-package.manifest")) as f:
                json.dump(manifest, f)
            underlay_user_link = os.path.join(
                underlay, ".click", "users", "@all", "test-package")
            overlay_user_link = os.path.join(
                overlay, ".click", "users", self.TEST_USER, "test-package")
            Click.ensuredir(os.path.dirname(underlay_user_link))
            os.symlink(underlay_unpacked, underlay_user_link)
            Click.ensuredir(os.path.dirname(overlay_user_link))
            os.symlink(overlay_unpacked, overlay_user_link)
            symlink_path = os.path.join(
                self.temp_dir, "test-package_test-app_1.0.test")
            underlay_target_path = os.path.join(underlay_user_link, "foo")
            overlay_target_path = os.path.join(overlay_user_link, "foo")
            os.symlink(overlay_target_path, symlink_path)
            hook = Click.Hook.open(db, "test")
            hook.sync(user_name=self.TEST_USER)
            self.assertTrue(os.path.islink(underlay_user_link))
            self.assertEqual(
                underlay_unpacked, os.readlink(underlay_user_link))
            self.assertFalse(os.path.islink(overlay_user_link))
            self.assertTrue(os.path.islink(symlink_path))
            self.assertEqual(underlay_target_path, os.readlink(symlink_path))


class TestPackageInstallHooks(TestClickHookBase):
    def test_removes_old_hooks(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            hooks_dir = os.path.join(self.temp_dir, "hooks")
            self._setup_hooks_dir(preloads, hooks_dir=hooks_dir)
            with mkfile(os.path.join(hooks_dir, "unity.hook")) as f:
                print("Pattern: %s/unity/${id}.scope" % self.temp_dir, file=f)
                print("Single-Version: yes", file=f)
            with mkfile(os.path.join(hooks_dir, "yelp-docs.hook")) as f:
                print("Pattern: %s/yelp/docs-${id}.txt" % self.temp_dir,
                      file=f)
                print("Single-Version: yes", file=f)
                print("Hook-Name: yelp", file=f)
            with mkfile(os.path.join(hooks_dir, "yelp-other.hook")) as f:
                print("Pattern: %s/yelp/other-${id}.txt" % self.temp_dir,
                      file=f)
                print("Single-Version: yes", file=f)
                print("Hook-Name: yelp", file=f)
            os.mkdir(os.path.join(self.temp_dir, "unity"))
            unity_path = os.path.join(
                self.temp_dir, "unity", "test_app_1.0.scope")
            os.symlink("dummy", unity_path)
            os.mkdir(os.path.join(self.temp_dir, "yelp"))
            yelp_docs_path = os.path.join(
                self.temp_dir, "yelp", "docs-test_app_1.0.txt")
            os.symlink("dummy", yelp_docs_path)
            yelp_other_path = os.path.join(
                self.temp_dir, "yelp", "other-test_app_1.0.txt")
            os.symlink("dummy", yelp_other_path)
            self._make_installed_click("test", "1.0", make_current=False, json_data={
                "hooks": {"app": {"yelp": "foo.txt", "unity": "foo.scope"}}})
            self._make_installed_click("test", "1.1", json_data={})
            Click.package_install_hooks(
                self.db, "test", "1.0", "1.1", user_name=None)
            self.assertFalse(os.path.lexists(unity_path))
            self.assertFalse(os.path.lexists(yelp_docs_path))
            self.assertFalse(os.path.lexists(yelp_other_path))

    def test_installs_new_hooks(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            hooks_dir = os.path.join(self.temp_dir, "hooks")
            self._setup_hooks_dir(preloads, hooks_dir=hooks_dir)
            with mkfile(os.path.join(hooks_dir, "a.hook")) as f:
                print("Pattern: %s/a/${id}.a" % self.temp_dir, file=f)
            with mkfile(os.path.join(hooks_dir, "b-1.hook")) as f:
                print("Pattern: %s/b/1-${id}.b" % self.temp_dir, file=f)
                print("Hook-Name: b", file=f)
            with mkfile(os.path.join(hooks_dir, "b-2.hook")) as f:
                print("Pattern: %s/b/2-${id}.b" % self.temp_dir, file=f)
                print("Hook-Name: b", file=f)
            os.mkdir(os.path.join(self.temp_dir, "a"))
            os.mkdir(os.path.join(self.temp_dir, "b"))
            self._make_installed_click(
                "test", "1.0", make_current=False, json_data={"hooks": {}})
            self._make_installed_click("test", "1.1", json_data={
                "hooks": {"app": {"a": "foo.a", "b": "foo.b"}}})
            Click.package_install_hooks(
                self.db, "test", "1.0", "1.1", user_name=None)
            self.assertTrue(os.path.lexists(
                os.path.join(self.temp_dir, "a", "test_app_1.1.a")))
            self.assertTrue(os.path.lexists(
                os.path.join(self.temp_dir, "b", "1-test_app_1.1.b")))
            self.assertTrue(os.path.lexists(
                os.path.join(self.temp_dir, "b", "2-test_app_1.1.b")))

    def test_upgrades_existing_hooks(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            hooks_dir = os.path.join(self.temp_dir, "hooks")
            self._setup_hooks_dir(preloads, hooks_dir=hooks_dir)
            with mkfile(os.path.join(hooks_dir, "a.hook")) as f:
                print("Pattern: %s/a/${id}.a" % self.temp_dir, file=f)
                print("Single-Version: yes", file=f)
            with mkfile(os.path.join(hooks_dir, "b-1.hook")) as f:
                print("Pattern: %s/b/1-${id}.b" % self.temp_dir, file=f)
                print("Single-Version: yes", file=f)
                print("Hook-Name: b", file=f)
            with mkfile(os.path.join(hooks_dir, "b-2.hook")) as f:
                print("Pattern: %s/b/2-${id}.b" % self.temp_dir, file=f)
                print("Single-Version: yes", file=f)
                print("Hook-Name: b", file=f)
            with mkfile(os.path.join(hooks_dir, "c.hook")) as f:
                print("Pattern: %s/c/${id}.c" % self.temp_dir, file=f)
                print("Single-Version: yes", file=f)
            os.mkdir(os.path.join(self.temp_dir, "a"))
            a_path = os.path.join(self.temp_dir, "a", "test_app_1.0.a")
            os.symlink("dummy", a_path)
            os.mkdir(os.path.join(self.temp_dir, "b"))
            b_irrelevant_path = os.path.join(
                self.temp_dir, "b", "1-test_other-app_1.0.b")
            os.symlink("dummy", b_irrelevant_path)
            b_1_path = os.path.join(self.temp_dir, "b", "1-test_app_1.0.b")
            os.symlink("dummy", b_1_path)
            b_2_path = os.path.join(self.temp_dir, "b", "2-test_app_1.0.b")
            os.symlink("dummy", b_2_path)
            os.mkdir(os.path.join(self.temp_dir, "c"))
            package_dir = os.path.join(self.temp_dir, "test")
            with mkfile(os.path.join(
                    package_dir, "1.0", ".click", "info",
                    "test.manifest")) as f:
                json.dump({"hooks": {"app": {"a": "foo.a", "b": "foo.b"}}}, f)
            with mkfile(os.path.join(
                    package_dir, "1.1", ".click", "info",
                    "test.manifest")) as f:
                json.dump(
                    {"hooks": {
                        "app": {"a": "foo.a", "b": "foo.b", "c": "foo.c"}}
                    }, f)
            Click.package_install_hooks(
                self.db, "test", "1.0", "1.1", user_name=None)
            self.assertFalse(os.path.lexists(a_path))
            self.assertTrue(os.path.lexists(b_irrelevant_path))
            self.assertFalse(os.path.lexists(b_1_path))
            self.assertFalse(os.path.lexists(b_2_path))
            self.assertTrue(os.path.lexists(
                os.path.join(self.temp_dir, "a", "test_app_1.1.a")))
            self.assertTrue(os.path.lexists(
                os.path.join(self.temp_dir, "b", "1-test_app_1.1.b")))
            self.assertTrue(os.path.lexists(
                os.path.join(self.temp_dir, "b", "2-test_app_1.1.b")))
            self.assertTrue(os.path.lexists(
                os.path.join(self.temp_dir, "c", "test_app_1.1.c")))


class TestPackageRemoveHooks(TestClickHookBase):
    def test_removes_hooks(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir") as (enter, preloads):
            enter()
            hooks_dir = os.path.join(self.temp_dir, "hooks")
            self._setup_hooks_dir(preloads, hooks_dir=hooks_dir)
            with mkfile(os.path.join(hooks_dir, "unity.hook")) as f:
                print("Pattern: %s/unity/${id}.scope" % self.temp_dir, file=f)
            with mkfile(os.path.join(hooks_dir, "yelp-docs.hook")) as f:
                print("Pattern: %s/yelp/docs-${id}.txt" % self.temp_dir,
                      file=f)
                print("Hook-Name: yelp", file=f)
            with mkfile(os.path.join(hooks_dir, "yelp-other.hook")) as f:
                print("Pattern: %s/yelp/other-${id}.txt" % self.temp_dir,
                      file=f)
                print("Hook-Name: yelp", file=f)
            os.mkdir(os.path.join(self.temp_dir, "unity"))
            unity_path = os.path.join(
                self.temp_dir, "unity", "test_app_1.0.scope")
            os.symlink("dummy", unity_path)
            os.mkdir(os.path.join(self.temp_dir, "yelp"))
            yelp_docs_path = os.path.join(
                self.temp_dir, "yelp", "docs-test_app_1.0.txt")
            os.symlink("dummy", yelp_docs_path)
            yelp_other_path = os.path.join(
                self.temp_dir, "yelp", "other-test_app_1.0.txt")
            os.symlink("dummy", yelp_other_path)
            package_dir = os.path.join(self.temp_dir, "test")
            with mkfile(os.path.join(
                    package_dir, "1.0", ".click", "info",
                    "test.manifest")) as f:
                json.dump(
                    {"hooks": {
                        "app": {"yelp": "foo.txt", "unity": "foo.scope"}}
                    }, f)
            Click.package_remove_hooks(self.db, "test", "1.0", user_name=None)
            self.assertFalse(os.path.lexists(unity_path))
            self.assertFalse(os.path.lexists(yelp_docs_path))
            self.assertFalse(os.path.lexists(yelp_other_path))


class TestPackageHooksValidateFramework(TestClickHookBase):

    def _setup_test_env(self, preloads):
        preloads["click_get_user_home"].return_value = b"/home/test-user"
        self._setup_hooks_dir(
            preloads, os.path.join(self.temp_dir, "hooks"))
        self._make_hook_file(dedent("""\
                           User-Level: yes
                           Pattern: %s/${id}.test
        """) % self.temp_dir)
        self.hook_symlink_path = os.path.join(
            self.temp_dir, "test-1_test1-app_1.0.test")

    def test_links_are_kept_on_validate_framework(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir", "click_get_user_home",
                "click_get_frameworks_dir",
                ) as (enter, preloads):
            enter()
            self._setup_frameworks(
                preloads, frameworks=["ubuntu-sdk-13.10"])
            self._setup_test_env(preloads)
            self._make_installed_click(json_data={
                "framework": "ubuntu-sdk-13.10",
                "hooks": {
                    "test1-app": {"test": "target-1"}
                },
            })
            self.assertTrue(os.path.lexists(self.hook_symlink_path))
            # run the hooks
            Click.run_user_hooks(self.db, user_name=self.TEST_USER)
            self.assertTrue(os.path.lexists(self.hook_symlink_path))

    def test_links_are_kept_multiple_frameworks(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir", "click_get_user_home",
                "click_get_frameworks_dir",
                ) as (enter, preloads):
            enter()
            self._setup_frameworks(
                preloads, frameworks=["ubuntu-sdk-14.04", "ubuntu-sdk-13.10"])
            self._setup_test_env(preloads)
            self._make_installed_click(json_data={
                "framework": "ubuntu-sdk-13.10",
                "hooks": {
                    "test1-app": {"test": "target-1"}
                },
            })
            self.assertTrue(os.path.lexists(self.hook_symlink_path))
            # run the hooks
            Click.run_user_hooks(self.db, user_name=self.TEST_USER)
            self.assertTrue(os.path.lexists(self.hook_symlink_path))

    def test_links_are_removed_on_missing_framework(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir", "click_get_user_home",
                "click_get_frameworks_dir",
                ) as (enter, preloads):
            enter()
            self._setup_frameworks(preloads, frameworks=["missing"])
            self._setup_test_env(preloads)
            self._make_installed_click(json_data={
                "framework": "ubuntu-sdk-13.10",
                "hooks": {
                    "test1-app": {"test": "target-1"}
                },
            })
            self.assertTrue(os.path.lexists(self.hook_symlink_path))
            # run the hooks
            Click.run_user_hooks(self.db, user_name=self.TEST_USER)
            self.assertFalse(os.path.lexists(self.hook_symlink_path))

    def test_links_are_removed_on_missing_multiple_framework(self):
        with self.run_in_subprocess(
                "click_get_hooks_dir", "click_get_user_home",
                "click_get_frameworks_dir",
                ) as (enter, preloads):
            enter()
            self._setup_frameworks(preloads, frameworks=["ubuntu-sdk-13.10"])
            self._setup_test_env(preloads)
            self._make_installed_click(json_data={
                "framework": "ubuntu-sdk-13.10, ubuntu-sdk-13.10-html",
                "hooks": {
                    "test1-app": {"test": "target-1"}
                },
            })
            self.assertTrue(os.path.lexists(self.hook_symlink_path))
            # run the hooks
            Click.run_user_hooks(self.db, user_name=self.TEST_USER)
            self.assertFalse(os.path.lexists(self.hook_symlink_path))



#! /usr/bin/python3

# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Verify a Click package."""

from __future__ import print_function

from optparse import OptionParser

from click_package.install import ClickInstaller


def run(argv):
    parser = OptionParser("%prog verify [options] PACKAGE-FILE")
    parser.add_option(
        "--force-missing-framework", action="store_true", default=False,
        help="ignore missing system framework")
    parser.add_option(
        "--allow-unauthenticated", action="store_true", default=False,
        help="allow installing packages with no sigantures")
    options, args = parser.parse_args(argv)
    if len(args) < 1:
        parser.error("need package file name")
    package_path = args[0]
    installer = ClickInstaller(
        db=None, force_missing_framework=options.force_missing_framework,
        allow_unauthenticated=options.allow_unauthenticated)
    installer.audit(package_path, slow=True)
    return 0



# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test that all top-level scripts work."""

__metaclass__ = type
__all__ = [
    'TestScripts',
    ]

import os
import subprocess
from unittest import skipIf

from click_package.tests.helpers import TestCase


class TestScripts(TestCase):
    @skipIf('SKIP_SLOW_TESTS' in os.environ, 'Skipping slow tests')
    def test_scripts(self):
        self.longMessage = True
        paths = []
        for dirpath, _, filenames in os.walk("bin"):
            filenames = [
                n for n in filenames
                if not n.startswith(".") and not n.endswith("~")]
            for filename in filenames:
                paths.append(os.path.join(dirpath, filename))
        for path in paths:
            subp = subprocess.Popen(
                [path, "--help"],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                universal_newlines=True)
            err = subp.communicate()[1]
            self.assertEqual("", err, "%s --help produced error output" % path)
            self.assertEqual(
                0, subp.returncode, "%s --help exited non-zero" % path)



# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file contains code from Python 3.3, released under the Python license
# (http://docs.python.org/3/license.html).

"""Testing helpers."""

from __future__ import print_function

__metaclass__ = type
__all__ = [
    'TestCase',
    'mkfile',
    'touch',
    ]


import contextlib
from functools import wraps
import json
import os
import re
import shutil
import sys
import tempfile
import unittest
try:
    from unittest import mock
except ImportError:
    import mock

from gi.repository import Click, GLib

from click_package.tests import gimock


def disable_logging(func):
    """Decorator to disable logging e.g. during a test"""
    @wraps(func)
    def wrapper(*args, **kwargs):
        import logging
        logging.disable(logging.CRITICAL)
        try:
            return func(*args, **kwargs)
        finally:
            logging.disable(logging.NOTSET)
    return wrapper


def make_installed_click(db, db_dir, package="test-1", version="1.0",
                         json_data={}, make_current=True, user="@all"):
    """Create a fake installed click package for the given db/db_dir"""
    json_data["name"] = package
    json_data["version"] = version
    with mkfile_utf8(os.path.join(
            db_dir, package, version, ".click", "info",
            "%s.manifest" % package)) as f:
        json.dump(json_data, f, ensure_ascii=False)
    if make_current:
        os.symlink(
            version, os.path.join(db_dir, package, "current"))
    if user == "@all":
        registry = Click.User.for_all_users(db)
    else:
        registry = Click.User.for_user(db, user)
    registry.set_version(package, version)
    return os.path.join(db_dir, package, version)


class TestCase(gimock.GIMockTestCase):
    def setUp(self):
        super(TestCase, self).setUp()
        self.temp_dir = None
        self.save_env = dict(os.environ)
        self.maxDiff = None

    def tearDown(self):
        for key in set(os.environ) - set(self.save_env):
            del os.environ[key]
        for key, value in os.environ.items():
            if value != self.save_env[key]:
                os.environ[key] = self.save_env[key]
        for key in set(self.save_env) - set(os.environ):
            os.environ[key] = self.save_env[key]

    def use_temp_dir(self):
        if self.temp_dir is not None:
            return self.temp_dir
        self.temp_dir = tempfile.mkdtemp(prefix="click")
        self.addCleanup(shutil.rmtree, self.temp_dir)
        return self.temp_dir

    # Monkey-patch for Python 2/3 compatibility.
    if not hasattr(unittest.TestCase, 'assertCountEqual'):
        assertCountEqual = unittest.TestCase.assertItemsEqual
    if not hasattr(unittest.TestCase, 'assertRegex'):
        assertRegex = unittest.TestCase.assertRegexpMatches
    # Renamed in Python 3.2 to omit the trailing 'p'.
    if not hasattr(unittest.TestCase, 'assertRaisesRegex'):
        assertRaisesRegex = unittest.TestCase.assertRaisesRegexp

    def assertRaisesGError(self, domain_name, code, callableObj,
                           *args, **kwargs):
        with self.assertRaises(GLib.GError) as cm:
            callableObj(*args, **kwargs)
        self.assertEqual(domain_name, cm.exception.domain)
        self.assertEqual(code, cm.exception.code)

    def assertRaisesFileError(self, code, callableObj, *args, **kwargs):
        self.assertRaisesGError(
            "g-file-error-quark", code, callableObj, *args, **kwargs)

    def assertRaisesDatabaseError(self, code, callableObj, *args, **kwargs):
        self.assertRaisesGError(
            "click_database_error-quark", code, callableObj, *args, **kwargs)

    def assertRaisesFrameworkError(self, code, callableObj, *args, **kwargs):
        self.assertRaisesGError(
            "click_framework_error-quark", code, callableObj, *args, **kwargs)

    def assertRaisesHooksError(self, code, callableObj, *args, **kwargs):
        self.assertRaisesGError(
            "click_hooks_error-quark", code, callableObj, *args, **kwargs)

    def assertRaisesQueryError(self, code, callableObj, *args, **kwargs):
        self.assertRaisesGError(
            "click_query_error-quark", code, callableObj, *args, **kwargs)

    def assertRaisesUserError(self, code, callableObj, *args, **kwargs):
        self.assertRaisesGError(
            "click_user_error-quark", code, callableObj, *args, **kwargs)

    def _setup_frameworks(self, preloads, frameworks_dir=None, frameworks=[]):
        frameworks_dir = self._create_mock_framework_dir(frameworks_dir)
        shutil.rmtree(frameworks_dir, ignore_errors=True)
        for framework in frameworks:
            self._create_mock_framework_file(framework)
        preloads["click_get_frameworks_dir"].side_effect = (
            lambda: self.make_string(frameworks_dir))

    def _create_mock_framework_dir(self, frameworks_dir=None):
        if frameworks_dir is None:
            frameworks_dir = os.path.join(self.temp_dir, "frameworks")
            patcher = mock.patch('click_package.paths.frameworks_dir', frameworks_dir)
            patcher.start()
            self.addCleanup(patcher.stop)
        Click.ensuredir(frameworks_dir)
        return frameworks_dir

    def _create_mock_framework_file(self, framework_name):
        self.use_temp_dir()
        self._create_mock_framework_dir()
        r = r'(?P<name>[a-z]+-sdk)-(?P<ver>[0-9.]+)(-[a-z0-9-]+)?'
        match = re.match(r, framework_name)
        if match is None:
            name = "unknown"
            ver = "1.0"
        else:
            name = match.group("name")
            ver = match.group("ver")
        framework_filename = os.path.join(
            self.temp_dir, "frameworks",
            "{0}.framework".format(framework_name))
        with open(framework_filename, "w") as f:
            f.write("Base-Name: {0}\n".format(name))
            f.write("Base-Version: {0}\n".format(ver))


if not hasattr(mock, "call"):
    # mock 0.7.2, the version in Ubuntu 12.04 LTS, lacks mock.ANY and
    # mock.call.  Since it's so convenient, monkey-patch a partial backport
    # (from Python 3.3 unittest.mock) into place here.
    class _ANY(object):
        "A helper object that compares equal to everything."

        def __eq__(self, other):
            return True

        def __ne__(self, other):
            return False

        def __repr__(self):
            return '<ANY>'

    mock.ANY = _ANY()

    class _Call(tuple):
        """
        A tuple for holding the results of a call to a mock, either in the form
        `(args, kwargs)` or `(name, args, kwargs)`.

        If args or kwargs are empty then a call tuple will compare equal to
        a tuple without those values. This makes comparisons less verbose::

            _Call(('name', (), {})) == ('name',)
            _Call(('name', (1,), {})) == ('name', (1,))
            _Call(((), {'a': 'b'})) == ({'a': 'b'},)

        The `_Call` object provides a useful shortcut for comparing with call::

            _Call(((1, 2), {'a': 3})) == call(1, 2, a=3)
            _Call(('foo', (1, 2), {'a': 3})) == call.foo(1, 2, a=3)

        If the _Call has no name then it will match any name.
        """
        def __new__(cls, value=(), name=None, parent=None, two=False,
                    from_kall=True):
            name = ''
            args = ()
            kwargs = {}
            _len = len(value)
            if _len == 3:
                name, args, kwargs = value
            elif _len == 2:
                first, second = value
                if isinstance(first, str):
                    name = first
                    if isinstance(second, tuple):
                        args = second
                    else:
                        kwargs = second
                else:
                    args, kwargs = first, second
            elif _len == 1:
                value, = value
                if isinstance(value, str):
                    name = value
                elif isinstance(value, tuple):
                    args = value
                else:
                    kwargs = value

            if two:
                return tuple.__new__(cls, (args, kwargs))

            return tuple.__new__(cls, (name, args, kwargs))

        def __init__(self, value=(), name=None, parent=None, two=False,
                     from_kall=True):
            self.name = name
            self.parent = parent
            self.from_kall = from_kall

        def __eq__(self, other):
            if other is mock.ANY:
                return True
            try:
                len_other = len(other)
            except TypeError:
                return False

            self_name = ''
            if len(self) == 2:
                self_args, self_kwargs = self
            else:
                self_name, self_args, self_kwargs = self

            other_name = ''
            if len_other == 0:
                other_args, other_kwargs = (), {}
            elif len_other == 3:
                other_name, other_args, other_kwargs = other
            elif len_other == 1:
                value, = other
                if isinstance(value, tuple):
                    other_args = value
                    other_kwargs = {}
                elif isinstance(value, str):
                    other_name = value
                    other_args, other_kwargs = (), {}
                else:
                    other_args = ()
                    other_kwargs = value
            else:
                # len 2
                # could be (name, args) or (name, kwargs) or (args, kwargs)
                first, second = other
                if isinstance(first, str):
                    other_name = first
                    if isinstance(second, tuple):
                        other_args, other_kwargs = second, {}
                    else:
                        other_args, other_kwargs = (), second
                else:
                    other_args, other_kwargs = first, second

            if self_name and other_name != self_name:
                return False

            # this order is important for ANY to work!
            return (other_args, other_kwargs) == (self_args, self_kwargs)

        def __ne__(self, other):
            return not self.__eq__(other)

        def __call__(self, *args, **kwargs):
            if self.name is None:
                return _Call(('', args, kwargs), name='()')

            name = self.name + '()'
            return _Call((self.name, args, kwargs), name=name, parent=self)

        def __getattr__(self, attr):
            if self.name is None:
                return _Call(name=attr, from_kall=False)
            name = '%s.%s' % (self.name, attr)
            return _Call(name=name, parent=self, from_kall=False)

    mock.call = _Call(from_kall=False)


@contextlib.contextmanager
def mkfile(path, mode="w"):
    Click.ensuredir(os.path.dirname(path))
    with open(path, mode) as f:
        yield f


@contextlib.contextmanager
def mkfile_utf8(path, mode="w"):
    Click.ensuredir(os.path.dirname(path))
    if sys.version < "3":
        import codecs
        with codecs.open(path, mode, "UTF-8") as f:
            yield f
    else:
        # io.open is available from Python 2.6, but we only use it with
        # Python 3 because it raises exceptions when passed bytes.
        import io
        with io.open(path, mode, encoding="UTF-8") as f:
            yield f


def touch(path):
    with mkfile(path, mode="a"):
        pass


def make_file_with_content(filename, content, mode=0o644):
    """Create a file with the given content and mode"""
    Click.ensuredir(os.path.dirname(filename))
    with open(filename, "w") as f:
        f.write(content)
    os.chmod(filename, mode)



# Copyright (C) 2014 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Helper functions to turn json-glib objects into Python objects."""

from __future__ import print_function

__metaclass__ = type
__all__ = [
    'ClickJsonError',
    'json_array_to_python',
    'json_node_to_python',
    'json_object_to_python',
    ]


from gi.repository import Json


class ClickJsonError(Exception):
    pass


def json_array_to_python(array):
    return [json_node_to_python(element) for element in array.get_elements()]


def json_object_to_python(obj):
    ret = {}
    for name in obj.get_members():
        ret[name] = json_node_to_python(obj.get_member(name))
    return ret


def json_node_to_python(node):
    node_type = node.get_node_type()
    if node_type == Json.NodeType.ARRAY:
        return json_array_to_python(node.get_array())
    elif node_type == Json.NodeType.OBJECT:
        return json_object_to_python(node.get_object())
    elif node_type == Json.NodeType.NULL:
        return None
    elif node_type == Json.NodeType.VALUE:
        return node.get_value()
    else:
        raise ClickJsonError(
            "Unknown JSON node type \"%s\"" % node_type.value_nick)



# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Click desktop hook.  (Temporary; do not rely on this.)"""

from __future__ import print_function

import errno
import io
import json
from optparse import OptionParser
import os

from gi.repository import Click

from click_package import osextras


COMMENT = \
    '# Generated by "click desktophook"; changes here will be overwritten.'


def desktop_entries(directory, only_ours=False):
    for entry in osextras.listdir_force(directory):
        if not entry.endswith(".desktop"):
            continue
        path = os.path.join(directory, entry)
        if only_ours:
            try:
                with io.open(path, encoding="UTF-8") as f:
                    if COMMENT not in f.read():
                        continue
            except Exception:
                continue
        yield entry


def split_entry(entry):
    entry = entry[:-8]  # strip .desktop
    return entry.split("_", 2)


def older(source_path, target_path):
    """Return True iff source_path is older than target_path.

    It's also OK for target_path to be missing.
    """
    try:
        source_mtime = os.stat(source_path).st_mtime
    except OSError as e:
        if e.errno == errno.ENOENT:
            return False
    try:
        target_mtime = os.stat(target_path).st_mtime
    except OSError as e:
        if e.errno == errno.ENOENT:
            return True
    return source_mtime < target_mtime


def read_hooks_for(path, package, app_name):
    try:
        directory = Click.find_package_directory(path)
        manifest_path = os.path.join(
            directory, ".click", "info", "%s.manifest" % package)
        with io.open(manifest_path, encoding="UTF-8") as manifest:
            return json.load(manifest).get("hooks", {}).get(app_name, {})
    except Exception:
        return {}


def quote_for_desktop_exec(s):
    """Quote a string for Exec in a .desktop file.

    The rules are fairly awful.  See:
      http://standards.freedesktop.org/desktop-entry-spec/latest/ar01s06.html
    """
    for c in s:
        if c in " \t\n\"'\\><~|&;$*?#()`%":
            break
    else:
        return s
    quoted = []
    for c in s:
        if c in "\"`$\\":
            quoted.append("\\" + c)
        elif c == "%":
            quoted.append("%%")
        else:
            quoted.append(c)
    escaped = []
    for c in "".join(quoted):
        if c == "\\":
            escaped.append("\\\\")
        else:
            escaped.append(c)
    return '"%s"' % "".join(escaped)


# TODO: This is a very crude .desktop file mangler; we should instead
# implement proper (de)serialisation.
def write_desktop_file(target_path, source_path, profile):
    Click.ensuredir(os.path.dirname(target_path))
    with io.open(source_path, encoding="UTF-8") as source, \
         io.open(target_path, "w", encoding="UTF-8") as target:
        source_dir = Click.find_package_directory(source_path)
        written_comment = False
        seen_path = False
        for line in source:
            if not line.rstrip("\n") or line.startswith("#"):
                # Comment
                target.write(line)
            elif line.startswith("["):
                # Group header
                target.write(line)
                if not written_comment:
                    print(COMMENT, file=target)
            elif "=" not in line:
                # Who knows?
                target.write(line)
            else:
                key, value = line.split("=", 1)
                key = key.strip()
                value = value.strip()
                if key == "Exec":
                    target.write(
                        "%s=aa-exec-click -p %s -- %s\n" %
                        (key, quote_for_desktop_exec(profile), value))
                elif key == "Path":
                    target.write("%s=%s\n" % (key, source_dir))
                    seen_path = True
                elif key == "Icon":
                    icon_path = os.path.join(source_dir, value)
                    if os.path.exists(icon_path):
                        target.write("%s=%s\n" % (key, icon_path))
                    else:
                        target.write("%s=%s\n" % (key, value))
                else:
                    target.write("%s=%s\n" % (key, value))
        if not seen_path:
            target.write("Path=%s\n" % source_dir)


def run(argv):
    parser = OptionParser("%prog desktophook [options]")
    parser.parse_args(argv)
    source_dir = os.path.expanduser("~/.local/share/click/hooks/desktop")
    target_dir = os.path.expanduser("~/.local/share/applications")
    source_entries = set(desktop_entries(source_dir))
    target_entries = set(desktop_entries(target_dir, only_ours=True))

    for new_entry in source_entries:
        package, app_name, version = split_entry(new_entry)
        source_path = os.path.join(source_dir, new_entry)
        target_path = os.path.join(target_dir, new_entry)
        if older(source_path, target_path):
            hooks = read_hooks_for(source_path, package, app_name)
            if "apparmor" in hooks:
                profile = "%s_%s_%s" % (package, app_name, version)
            else:
                profile = "unconfined"
            write_desktop_file(target_path, source_path, profile)

    for remove_entry in target_entries - source_entries:
        os.unlink(os.path.join(target_dir, remove_entry))

    return 0



# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for click.paths.

This is mostly just to reduce noise in the coverage report.
"""

from __future__ import print_function

__metaclass__ = type
__all__ = [
    'TestClickPaths',
    ]


from gi.repository import Click

from click_package.tests.helpers import TestCase


class TestClickPaths(TestCase):
    def test_get_hooks_dir(self):
        self.assertEqual("/usr/share/click/hooks", Click.get_hooks_dir())

    def test_get_db_dir(self):
        self.assertEqual("/usr/etc/click/databases", Click.get_db_dir())

    def test_get_frameworks_dir(self):
        self.assertEqual("/usr/share/click/frameworks", Click.get_frameworks_dir())



# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt <mvo@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""List available frameworks."""

from __future__ import print_function

from argparse import ArgumentParser

from gi.repository import Click


def list(parser, args):
    for framework in Click.Framework.get_frameworks():
        print("%s" % framework.props.name)
    return 0


def info(parser, args):
    framework = Click.Framework.open(args.framework_name)
    for field in sorted(framework.get_fields()):
        print("%s: %s" % (field, framework.get_field(field)))


def get_field(parser, args):
    framework = Click.Framework.open(args.framework_name)
    print(framework.get_field(args.field_name))


def run(argv):
    parser = ArgumentParser("click framework")
    subparsers = parser.add_subparsers()
    list_parser = subparsers.add_parser(
        "list",
        help="list available frameworks")
    list_parser.set_defaults(func=list)
    info_parser = subparsers.add_parser(
        "info",
        help="show info about a specific framework")
    info_parser.add_argument(
        "framework_name",
        help="framework name with the information")
    info_parser.set_defaults(func=info)
    get_field_parser = subparsers.add_parser(
        "get-field",
        help="get a field from a given framework")
    get_field_parser.add_argument(
        "framework_name",
        help="framework name with the information")
    get_field_parser.add_argument(
        "field_name",
        help="the field name (e.g. base-version)")
    get_field_parser.set_defaults(func=get_field)

    args = parser.parse_args(argv)
    if not hasattr(args, "func"):
        parser.print_help()
        return 1
    return args.func(parser, args)



# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt <michael.vogt@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Integration tests for the click CLI verify command."""

import os
import subprocess

from .helpers import ClickTestCase


class TestVerify(ClickTestCase):
    def test_verify_force_missing_framework_ok(self):
        name = "com.example.verify-missing-framework"
        path_to_click = self._make_click(name)
        output = subprocess.check_output([
            self.click_binary, "verify",
            "--force-missing-framework",
            "--allow-unauthenticated",
            path_to_click], universal_newlines=True)
        self.assertEqual(output, "")

    def test_verify_force_ok(self):
        name = "com.example.verify-ok"
        path_to_click = self._make_click(name, framework="")
        output = subprocess.check_output([
            self.click_binary, "verify", "--allow-unauthenticated",
            path_to_click], universal_newlines=True)
        self.assertEqual(output, "")

    def test_verify_missing_framework(self):
        name = "com.example.verify-really-missing-framework"
        path_to_click = self._make_click(name, framework="missing")
        with self.assertRaises(subprocess.CalledProcessError) as cm:
            subprocess.check_output(
                [self.click_binary, "verify",
                 "--allow-unauthenticated",
                 path_to_click],
                universal_newlines=True, stderr=subprocess.STDOUT)
        expected_error = (
            'click_package.framework.ClickFrameworkInvalid: Framework '
            '"missing" not present on system (use '
            '--force-missing-framework option to override)')
        self.assertIn(expected_error, cm.exception.output)

    def test_verify_no_click_but_invalid(self):
        name = "com.example.verify-no-click"
        path_to_click = os.path.join(self.temp_dir, name+".click")
        with open(path_to_click, "w") as f:
            f.write("something-that-is-not-a-click")
        with self.assertRaises(subprocess.CalledProcessError) as cm:
            subprocess.check_output(
                [self.click_binary, "verify", "--allow-unauthenticated",
                 path_to_click],
                universal_newlines=True, stderr=subprocess.STDOUT)
        expected_error = (
            'click_package.install.DebsigVerifyError: '
            'Signature verification error: '
            'debsig: %s does not appear to be a deb format package'
        ) % path_to_click
        self.assertIn(expected_error, cm.exception.output)



# Copyright (C) 2014 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Mock function support based on GObject Introspection.

(Note to reviewers: I expect to rewrite this from scratch on my own time as
a more generalised set of Python modules for unit testing of C code,
although using similar core ideas.  This is a first draft for the purpose of
getting Click's test suite to work expediently, rather than an interface I'm
prepared to commit to long-term.)

Python is a versatile and concise language for writing tests, and GObject
Introspection (GI) makes it straightforward (often trivial) to bind native
code into Python.  However, writing tests for native code quickly runs into
the problem of how to build mock functions.  You might reasonably have code
that calls chown(), for instance, and want to test how it's called rather
than worrying about setting up a fakeroot-type environment where chown()
will work.  The obvious solution is to use `LD_PRELOAD` wrappers, but there
are various problems to overcome in practice:

 * You can only set up a new `LD_PRELOAD` by going through the run-time
   linker; you can't just set it for a single in-process test case.
 * Generating the preloaded wrapper involves a fair bit of boilerplate code.
 * Having to write per-test mock code in C is inconvenient, and makes it
   difficult to get information back out of the mock (such as "how often was
   this function called, and with what arguments?").

The first problem can be solved by a decorator that knows how to run
individual tests in a subprocess.  This is made somewhat more inconvenient
by the fact that there is no way for a context manager's `__enter__` method
to avoid executing the context-managed block other than by throwing an
exception, which makes it hard to silently avoid executing the test case in
the parent process, but we can work around this at the cost of an extra line
of code per invocation.

For the rest, a combination of GI itself and ctypes can help.  We can use GI
to keep track of argument and return types of the mocked C functions in a
reasonably sane way, by parsing header files.  We're operating in the other
direction from how GI is normally used, so PyGObject can't deal with
bridging the two calling conventions for us.  ctypes can: but we still need
to be careful!  We have to construct the callback functions in the child
process, ensure that we keep references to them, and inject function
pointers into the preloaded library via specially-named helper functions;
until those function pointers are set up we must make sure to call the libc
functions instead (since some of them might be called during Python
startup).

The combination of all of this allows us to bridge C functions somewhat
transparently into Python.  This lets you supply a Python function or method
as the mock replacement for a C library function, making it much simpler to
record state.

It's still not perfect:

 * We're using GI in an upside-down kind of way, and we specifically need
   GIR files rather than typelibs so that we can extract the original C
   type, so some fiddling is required for each new function you want to
   mock.

 * The subprocess arrangements are unavoidably slow and it's possible that
   they may cause problems with some test runners.

 * Some C functions (such as `stat`) tend to have multiple underlying entry
   points in the C library which must be preloaded independently.

 * You have to be careful about how your libraries are linked, because `ld
   -Wl,-Bsymbolic-functions` prevents `LD_PRELOAD` working for intra-library
   calls.

 * `ctypes should return composite types from callbacks
   <http://bugs.python.org/issue5710>`_.  The least awful approach for now
   seems to be to construct the composite type in question, stash a
   reference to it forever, and then return a pointer to it as a void *; we
   can only get away with this because tests are by nature relatively
   short-lived.

 * The ctypes module's handling of 64-bit pointers is basically just awful.
   The right answer is probably to use a different callback-generation
   framework entirely (maybe extending PyGObject so that we can get at the
   pieces we need), but I've hacked around it for now.

 * It doesn't appear to be possible to install mock replacements for
   functions that are called directly from Python code using their GI
   wrappers.  You can work around this by simply patching the GI wrapper
   instead, using `mock.patch`.

I think the benefits, in terms of local clarity of tests, are worth the
downsides.
"""

from __future__ import print_function

__metaclass__ = type
__all__ = ['GIMockTestCase']


import contextlib
import ctypes
import fcntl
from functools import partial
import os
import pickle
import shutil
import subprocess
import sys
import tempfile
from textwrap import dedent
import traceback
import unittest
try:
    from unittest import mock
except ImportError:
    import mock
try:
    import xml.etree.cElementTree as etree
except ImportError:
    import xml.etree.ElementTree as etree

from click_package.tests.gimock_types import Stat, Stat64
from click_package.tests import config, get_executable

# Borrowed from giscanner.girparser.
CORE_NS = "http://www.gtk.org/introspection/core/1.0"
C_NS = "http://www.gtk.org/introspection/c/1.0"
GLIB_NS = "http://www.gtk.org/introspection/glib/1.0"


def _corens(tag):
    return '{%s}%s' % (CORE_NS, tag)


def _glibns(tag):
    return '{%s}%s' % (GLIB_NS, tag)


def _cns(tag):
    return '{%s}%s' % (C_NS, tag)


# Override some c:type annotations that g-ir-scanner gets a bit wrong.
_c_type_override = {
    "passwd*": "struct passwd*",
    "stat*": "struct stat*",
    "stat64*": "struct stat64*",
    }


# Mapping of GI type name -> ctypes type.
_typemap = {
    "GError**": ctypes.c_void_p,
    "gboolean": ctypes.c_int,
    "gint": ctypes.c_int,
    "gint*": ctypes.POINTER(ctypes.c_int),
    "gint32": ctypes.c_int32,
    "gpointer": ctypes.c_void_p,
    "guint": ctypes.c_uint,
    "guint8**": ctypes.POINTER(ctypes.POINTER(ctypes.c_uint8)),
    "guint32": ctypes.c_uint32,
    "none": None,
    "utf8": ctypes.c_char_p,
    "utf8*": ctypes.POINTER(ctypes.c_char_p),
    }


class GIMockTestCase(unittest.TestCase):
    def setUp(self):
        super(GIMockTestCase, self).setUp()
        self._preload_func_refs = []
        self._composite_refs = []
        self._delegate_funcs = {}

    def tearDown(self):
        self._preload_func_refs = []
        self._composite_refs = []
        self._delegate_funcs = {}

    def doCleanups(self):
        # we do not want to run the cleanups twice, just run it in the parent
        if "GIMOCK_SUBPROCESS" not in os.environ:
            return super(GIMockTestCase, self).doCleanups()

    def _gir_get_type(self, obj):
        ret = {}
        arrayinfo = obj.find(_corens("array"))
        if arrayinfo is not None:
            typeinfo = arrayinfo.find(_corens("type"))
            raw_ctype = arrayinfo.get(_cns("type"))
        else:
            typeinfo = obj.find(_corens("type"))
            raw_ctype = typeinfo.get(_cns("type"))
        gi_type = typeinfo.get("name")
        if obj.get("direction", "in") == "out":
            gi_type += "*"
        if arrayinfo is not None:
            gi_type += "*"
        ret["gi"] = gi_type
        ret["c"] = _c_type_override.get(raw_ctype, raw_ctype)
        return ret

    def _parse_gir(self, path):
        # A very, very crude GIR parser.  We might have used
        # giscanner.girparser, but it's not importable in Python 3 at the
        # moment.
        tree = etree.parse(path)
        root = tree.getroot()
        assert root.tag == _corens("repository")
        assert root.get("version") == "1.2"
        ns = root.find(_corens("namespace"))
        assert ns is not None
        funcs = {}
        for func in ns.findall(_corens("function")):
            name = func.get(_cns("identifier"))
            # g-ir-scanner skips identifiers starting with "__", which we
            # need in order to mock stat effectively.  Work around this.
            name = name.replace("under_under_", "__")
            headers = None
            for attr in func.findall(_corens("attribute")):
                if attr.get("name") == "headers":
                    headers = attr.get("value")
                    break
            rv = func.find(_corens("return-value"))
            assert rv is not None
            params = []
            paramnode = func.find(_corens("parameters"))
            if paramnode is not None:
                for param in paramnode.findall(_corens("parameter")):
                    params.append({
                        "name": param.get("name"),
                        "type": self._gir_get_type(param),
                        })
            if func.get("throws", "0") == "1":
                params.append({
                    "name": "error",
                    "type": {"gi": "GError**", "c": "GError**"},
                    })
            funcs[name] = {
                "name": name,
                "headers": headers,
                "rv": self._gir_get_type(rv),
                "params": params,
                }
        return funcs

    def _ctypes_type(self, gi_type):
        return _typemap[gi_type["gi"]]

    def make_preloads(self, preloads):
        rpreloads = []
        std_headers = set([
            "dlfcn.h",
            # Not strictly needed, but convenient for ad-hoc debugging.
            "stdio.h",
            "stdint.h",
            "stdlib.h",
            "string.h",
            "sys/types.h",
            "unistd.h",
            ])
        preload_headers = set()
        funcs = self._parse_gir("click_package/tests/preload.gir")
        for name, func in preloads.items():
            info = funcs[name]
            rpreloads.append([info, func])
            headers = info["headers"]
            if headers is not None:
                preload_headers.update(headers.split(","))
        if "GIMOCK_SUBPROCESS" in os.environ:
            return None, rpreloads
        self._gimock_temp_dir = tempfile.mkdtemp(prefix="gimock")
        self.addCleanup(shutil.rmtree, self._gimock_temp_dir)
        preloads_dir = os.path.join(self._gimock_temp_dir, "_preloads")
        os.makedirs(preloads_dir)
        c_path = os.path.join(preloads_dir, "gimockpreload.c")
        with open(c_path, "w") as c:
            print("#define _GNU_SOURCE", file=c)
            for header in sorted(std_headers | preload_headers):
                print("#include <%s>" % header, file=c)
            print(file=c)
            for info, _ in rpreloads:
                conv = {}
                conv["name"] = info["name"]
                argtypes = [p["type"]["c"] for p in info["params"]]
                argnames = [p["name"] for p in info["params"]]
                conv["ret"] = info["rv"]["c"]
                conv["bareproto"] = ", ".join(argtypes)
                conv["proto"] = ", ".join(
                    "%s %s" % pair for pair in zip(argtypes, argnames))
                conv["args"] = ", ".join(argnames)
                if conv["ret"] == "gchar*":
                    conv["need_strdup"] = "strdup"
                else:
                    conv["need_strdup"] = ""
                # The delegation scheme used here is needed because trying
                # to pass pointers back and forward through ctypes is a
                # recipe for having them truncated to 32 bits at the drop of
                # a hat.  This approach is less obvious but much safer.
                print(dedent("""\
                    typedef %(ret)s preloadtype_%(name)s (%(bareproto)s);
                    preloadtype_%(name)s *ctypes_%(name)s = (void *) 0;
                    preloadtype_%(name)s *real_%(name)s = (void *) 0;
                    static volatile int delegate_%(name)s = 0;

                    extern void _gimock_init_%(name)s (preloadtype_%(name)s *f)
                    {
                        ctypes_%(name)s = f;
                        if (! real_%(name)s) {
                            /* Retry lookup in case the symbol wasn't
                             * resolvable until the program under test was
                             * loaded.
                             */
                            char *err;
                            dlerror ();
                            real_%(name)s = dlsym (RTLD_NEXT, "%(name)s");
                            if ((err = dlerror ()) != NULL) {
                                fprintf (stderr,
                                         "Error getting address of symbol "
                                         "'%(name)s': %%s\\n", err);
                                fflush (stderr);
                                _exit (1);
                            }
                        }
                    }
                    """) % conv, file=c)
                if conv["ret"] == "void":
                    print(dedent("""\
                        void %(name)s (%(proto)s)
                        {
                            if (ctypes_%(name)s) {
                                delegate_%(name)s = 0;
                                (*ctypes_%(name)s) (%(args)s);
                                if (! delegate_%(name)s)
                                    return;
                            }
                            if (!real_%(name)s)
                                _gimock_init_%(name)s (ctypes_%(name)s);
                            (*real_%(name)s) (%(args)s);
                        }
                        """) % conv, file=c)
                else:
                    print(dedent("""\
                        %(ret)s %(name)s (%(proto)s)
                        {
                            if (ctypes_%(name)s) {
                                %(ret)s ret;
                                delegate_%(name)s = 0;
                                ret = (*ctypes_%(name)s) (%(args)s);
                                if (! delegate_%(name)s)
                                    return %(need_strdup)s(ret);
                            }
                            if (!real_%(name)s)
                                _gimock_init_%(name)s (ctypes_%(name)s);
                            return (*real_%(name)s) (%(args)s);
                        }
                        """) % conv, file=c)
                print(dedent("""\
                    extern void _gimock_delegate_%(name)s (void)
                    {
                        delegate_%(name)s = 1;
                    }
                    """) % conv, file=c)
            print(dedent("""\
                static void __attribute__ ((constructor))
                gimockpreload_init (void)
                {
                    char *err;
                    dlerror ();
                """), file=c)
            print("\n".join("    " + line for line in (dedent("""\
                    real_%(name)s = dlsym (RTLD_NEXT, "%(name)s");
                    if ((err = dlerror ()) != NULL) {
                        fprintf (stderr,
                                 "Error getting address of symbol "
                                 "'%(name)s': %%s\\n", err);
                        fflush (stderr);
                        _exit (1);
                    }
                """) % {"name": name}).split("\n")), file=c)
            print("}", file=c)
        if "GIMOCK_PRELOAD_DEBUG" in os.environ:
            with open(c_path) as c:
                print(c.read())
        # TODO: Use libtool or similar rather than hardcoding gcc invocation.
        lib_path = os.path.join(preloads_dir, "libgimockpreload.so")
        libraries = ["glib-2.0", "gee-0.8", "json-glib-1.0"]
        cflags = subprocess.check_output(
            ["pkg-config", "--cflags"] + libraries,
            universal_newlines=True).rstrip("\n").split()
        libs = subprocess.check_output(
            ["pkg-config", "--libs"] + libraries,
            universal_newlines=True).rstrip("\n").split()
        compile_cmd = [
            "gcc", "-O0", "-g", "-shared", "-fPIC", "-DPIC",
            "-I", "lib/click",
            "-L",
            os.path.join(config.abs_top_builddir, "lib", "click", ".libs"),
            ]
        compile_cmd.extend(cflags)
        compile_cmd.extend(["-Wl,-soname", "-Wl,libgimockpreload.so"])
        compile_cmd.append("-Wl,--no-as-needed")
        compile_cmd.append(c_path)
        compile_cmd.append("-lclick-0.4")
        compile_cmd.extend(libs)
        compile_cmd.append("-ldl")
        compile_cmd.extend(["-o", lib_path])
        subprocess.check_call(compile_cmd)
        return lib_path, rpreloads

    # Use as:
    #   with self.run_in_subprocess("func", ...) as (enter, preloads):
    #       enter()
    #       # test case body; preloads["func"] will be a mock.MagicMock
    #       # instance
    @contextlib.contextmanager
    def run_in_subprocess(self, *patches):
        preloads = {}
        for patch in patches:
            preloads[patch] = mock.MagicMock()
        if preloads:
            lib_path, rpreloads = self.make_preloads(preloads)
        else:
            lib_path, rpreloads = None, None

        class ParentProcess(Exception):
            pass

        def helper(lib_path, rpreloads):
            if "GIMOCK_SUBPROCESS" in os.environ:
                del os.environ["LD_PRELOAD"]
                preload_lib = ctypes.cdll.LoadLibrary(lib_path)
                delegate_cfunctype = ctypes.CFUNCTYPE(None)
                for info, func in rpreloads:
                    signature = [info["rv"]] + [
                        p["type"] for p in info["params"]]
                    signature = [self._ctypes_type(t) for t in signature]
                    cfunctype = ctypes.CFUNCTYPE(*signature)
                    init = getattr(
                        preload_lib, "_gimock_init_%s" % info["name"])
                    cfunc = cfunctype(func)
                    self._preload_func_refs.append(cfunc)
                    init(cfunc)
                    delegate = getattr(
                        preload_lib, "_gimock_delegate_%s" % info["name"])
                    self._delegate_funcs[info["name"]] = delegate_cfunctype(
                        delegate)
                return
            rfd, wfd = os.pipe()
            # It would be cleaner to use subprocess.Popen(pass_fds=[wfd]), but
            # that isn't available in Python 2.7.
            if hasattr(os, "set_inheritable"):
                os.set_inheritable(wfd, True)
            else:
                fcntl.fcntl(rfd, fcntl.F_SETFD, fcntl.FD_CLOEXEC)
            args = get_executable() + [
                "-m", "unittest",
                "%s.%s.%s" % (
                    self.__class__.__module__, self.__class__.__name__,
                    self._testMethodName)]
            env = os.environ.copy()
            env["GIMOCK_SUBPROCESS"] = str(wfd)
            if lib_path is not None:
                env["LD_PRELOAD"] = lib_path
            subp = subprocess.Popen(args, close_fds=False, env=env,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
            os.close(wfd)
            reader = os.fdopen(rfd, "rb")
            stdout, stderr = subp.communicate()
            exctype = pickle.load(reader)
            # "normal" exit
            if exctype is not None and issubclass(exctype, SystemExit):
                pass
            elif exctype is not None and issubclass(exctype, AssertionError):
                print(stdout, file=sys.stdout)
                print(stderr, file=sys.stderr)
                raise AssertionError("Subprocess failed a test!")
            elif exctype is not None or subp.returncode != 0:
                print(stdout, file=sys.stdout)
                print(stderr, file=sys.stderr)
                raise Exception("Subprocess returned an error! (%s, %s)" % (
                    exctype, subp.returncode))
            reader.close()
            raise ParentProcess()

        try:
            yield partial(helper, lib_path, rpreloads), preloads
            if "GIMOCK_SUBPROCESS" in os.environ:
                wfd = int(os.environ["GIMOCK_SUBPROCESS"])
                writer = os.fdopen(wfd, "wb")
                # a sys.ext will generate a SystemExit exception, so we
                # push it into the pipe so that the parent knows whats going on
                pickle.dump(SystemExit, writer)
                writer.flush()
                sys.exit(0)
        except ParentProcess:
            pass
        except Exception as e:
            if "GIMOCK_SUBPROCESS" in os.environ:
                wfd = int(os.environ["GIMOCK_SUBPROCESS"])
                writer = os.fdopen(wfd, "wb")
                # It would be better to use tblib to pickle the traceback so
                # that we can re-raise it properly from the parent process.
                # Until that's packaged and available to us, just print the
                # traceback and send the exception type.
                print()
                traceback.print_exc()
                pickle.dump(type(e), writer)
                writer.flush()
                os._exit(1)
            else:
                raise

    def make_pointer(self, composite):
        # Store a reference to a composite type and return a pointer to it,
        # working around http://bugs.python.org/issue5710.
        self._composite_refs.append(composite)
        return ctypes.addressof(composite)

    def make_string(self, s):
        # As make_pointer, but for a string.
        copied = ctypes.create_string_buffer(s.encode())
        self._composite_refs.append(copied)
        return ctypes.addressof(copied)

    def convert_pointer(self, composite_type, address):
        # Return a ctypes composite type instance at a given address.
        return composite_type.from_address(address)

    def convert_stat_pointer(self, name, address):
        # As convert_pointer, but for a "struct stat *" or "struct stat64 *"
        # depending on the wrapped function name.
        stat_type = {"__xstat": Stat, "__xstat64": Stat64}
        return self.convert_pointer(stat_type[name], address)

    def delegate_to_original(self, name):
        # Cause the wrapper function to delegate to the original version
        # after the callback returns.  (Note that the callback still needs
        # to return something type-compatible with the declared result type,
        # although the return value will otherwise be ignored.)
        self._delegate_funcs[name]()



# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt <michael.vogt@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Integration tests for the click hook feature."""

import os
import subprocess
from textwrap import dedent

from .helpers import (
    ClickTestCase,
    require_root,
)


class TestHook(ClickTestCase):

    @classmethod
    def setUpClass(cls):
        super(TestHook, cls).setUpClass()
        require_root()

    def _make_hook(self, name):
        hook_fname = "/usr/share/click/hooks/%s.hook" % name
        canary_fname = os.path.join(self.temp_dir, "canary.sh")
        canary_log = os.path.join(self.temp_dir, "canary.log")
        with open(hook_fname, "w") as f:
            f.write(dedent("""\
            Pattern: ${home}/${id}.test-hook
            User-Level: yes
            Exec: %s
            Hook-Name: %s
            """ % (canary_fname, name)))
        with open(canary_fname, "w") as f:
            f.write(dedent("""\
            #!/bin/sh
            echo "i-hook-you-up" >> %s
            """ % canary_log))
        os.chmod(canary_fname, 0o755)
        return hook_fname, canary_log

    def test_hook_install_user(self):
        # build/install the hook
        hook_name = "clicktesthook"
        hook_file, hook_log = self._make_hook(hook_name)
        self.addCleanup(os.unlink, hook_file)
        subprocess.check_call(
            [self.click_binary, "hook", "install", hook_name])
        self.addCleanup(
            subprocess.check_call, [self.click_binary, "hook", "remove",
                                    hook_name])
        # make click that uses the hook
        hooks = {'app1': {hook_name: 'README'}}
        click_pkg_name = "com.example.hook-1"
        click_pkg = self._make_click(
            click_pkg_name, framework="", hooks=hooks)
        user = os.environ.get("USER", "root")
        self.click_install(click_pkg, click_pkg_name, user)
        # ensure we have the hook
        generated_hook_file = os.path.expanduser(
            "~/com.example.hook-1_app1_1.0.test-hook")
        self.assertTrue(os.path.exists(generated_hook_file))
        self.assertEqual(
            os.path.realpath(generated_hook_file),
            "/opt/click.ubuntu.com/com.example.hook-1/1.0/README")
        with open(hook_log) as f:
            hook_log_content = f.read().strip()
        self.assertEqual("i-hook-you-up", hook_log_content)



# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Show manifest information for a Click package."""

from __future__ import print_function

from contextlib import closing
import glob
import json
from optparse import OptionParser
import os
import sys

from gi.repository import Click

from click_package.install import DebFile
from click_package.json_helpers import json_object_to_python


def _load_manifest(manifest_file):
    manifest = json.load(manifest_file)
    keys = list(manifest)
    for key in keys:
        if key.startswith("_"):
            del manifest[key]
    return manifest


def get_manifest(options, arg):
    if "/" not in arg:
        db = Click.DB()
        db.read(db_dir=None)
        if options.root is not None:
            db.add(options.root)
        registry = Click.User.for_user(db, name=options.user)
        if registry.has_package_name(arg):
            return json_object_to_python(registry.get_manifest(arg))

    try:
        with closing(DebFile(filename=arg)) as package:
            with package.control.get_file(
                    "manifest", encoding="UTF-8") as manifest_file:
                return _load_manifest(manifest_file)
    except Exception:
        pkgdir = Click.find_package_directory(arg)
        manifest_path = glob.glob(
            os.path.join(pkgdir, ".click", "info", "*.manifest"))
        if len(manifest_path) > 1:
            raise Exception("Multiple manifest files found in '%s'" % (
                manifest_path))
        with open(manifest_path[0]) as f:
            return _load_manifest(f)


def run(argv):
    parser = OptionParser("%prog info [options] PATH")
    parser.add_option(
        "--root", metavar="PATH", help="look for additional packages in PATH")
    parser.add_option(
        "--user", metavar="USER",
        help="look up PACKAGE-NAME for USER (if you have permission; "
             "default: current user)")
    options, args = parser.parse_args(argv)
    if len(args) < 1:
        parser.error("need file name")
    try:
        manifest = get_manifest(options, args[0])
    except Exception as e:
        print(e, file=sys.stderr)
        return 1
    json.dump(
        manifest, sys.stdout, ensure_ascii=False, sort_keys=True, indent=4,
        separators=(",", ": "))
    print()
    return 0



# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test compliance with various static analysis tools."""

from __future__ import print_function

__metaclass__ = type
__all__ = [
    'TestStatic',
    ]


import os
import sys
from unittest import skipIf

from pkg_resources import resource_filename

try:
    import pep8
except ImportError:
    pep8 = None
try:
    import pyflakes
    import pyflakes.api
    import pyflakes.reporter
except ImportError:
    pyflakes = None


from click_package.tests.helpers import TestCase


class TestStatic(TestCase):
    def all_paths(self):
        paths = []
        start_dir = os.path.dirname(resource_filename('click_package', '__init__.py'))
        for dirpath, dirnames, filenames in os.walk(start_dir):
            for ignore in ('doc', ".bzr", "__pycache__"):
                if ignore in dirnames:
                    dirnames.remove(ignore)
            filenames = [
                n for n in filenames
                if not n.startswith(".") and not n.endswith("~")]
            if dirpath.split(os.sep)[-1] == "bin":
                for filename in filenames:
                    paths.append(os.path.join(dirpath, filename))
            else:
                for filename in filenames:
                    if filename.endswith(".py"):
                        paths.append(os.path.join(dirpath, filename))
        return paths

    @skipIf('SKIP_SLOW_TESTS' in os.environ, 'Skipping slow tests')
    @skipIf(pep8 is None, 'No pep8 package available')
    def test_pep8_clean(self):
        # https://github.com/jcrocholl/pep8/issues/103
        pep8_style = pep8.StyleGuide(ignore='E123')
        result = pep8_style.check_files(self.all_paths())
        self.assertEqual(result.total_errors, 0)

    @skipIf('SKIP_SLOW_TESTS' in os.environ, 'Skipping slow tests')
    @skipIf(pyflakes is None, 'No pyflakes package available')
    def test_pyflakes_clean(self):
        reporter = pyflakes.reporter.Reporter(sys.stdout, sys.stderr)
        warnings = pyflakes.api.checkRecursive(self.all_paths(), reporter)
        self.assertEqual(0, warnings)



# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt <michael.vogt@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Integration tests helper for the click CLI interface."""

import contextlib
import glob
import json
import os
import random
import shutil
import string
import subprocess
import tempfile
import unittest


def require_root():
    if os.getuid() != 0:
        raise unittest.SkipTest("This test needs to run as root")


def require_network():
    try:
        if subprocess.call(["ping", "-c1", "archive.ubuntu.com"]) != 0:
            raise unittest.SkipTest("Need network")
    except Exception:
        pass


def require_overlay():
    try:
        subprocess.check_call(["/sbin/modprobe", "overlay"])
    except subprocess.CalledProcessError:
        raise unittest.SkipTest("Requires overlay fs support")


@contextlib.contextmanager
def chdir(target):
    curdir = os.getcwd()
    os.chdir(target)
    try:
        yield
    finally:
        os.chdir(curdir)


def cmdline_for_user(username):
    """Helper to get the click commandline for the given username"""
    if username == "@all":
        user = "--all-users"
    else:
        user = "--user=%s" % username
    return user


class ClickTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        if "TEST_INTEGRATION" not in os.environ:
            raise unittest.SkipTest("Skipping integration tests")
        cls.click_binary = os.environ.get("CLICK_BINARY", "/usr/bin/click")

    def setUp(self):
        super(ClickTestCase, self).setUp()
        self.temp_dir = tempfile.mkdtemp()

    def tearDown(self):
        super(ClickTestCase, self).tearDown()
        # we force the cleanup before removing the tempdir so that stuff
        # in temp_dir is still available
        self.doCleanups()
        shutil.rmtree(self.temp_dir)

    def click_install(self, path_to_click, click_name, username,
                      allow_unauthenticated=True):
        cmd = [self.click_binary, "install", cmdline_for_user(username)]
        if allow_unauthenticated:
            cmd.append("--allow-unauthenticated")
        cmd.append(path_to_click)
        subprocess.check_call(cmd)
        self.addCleanup(self.click_unregister, click_name, username)

    def click_unregister(self, click_name, username):
        subprocess.check_call(
            [self.click_binary, "unregister", cmdline_for_user(username),
             click_name])

    def _create_manifest(self, target, name, version, framework, hooks={}):
        with open(target, "w") as f:
            json.dump({
                'name': name,
                'version': str(version),
                'maintainer': 'Foo Bar <foo@example.org>',
                'title': 'test title',
                'framework': framework,
                'hooks': hooks,
                }, f)

    def _make_click(self, name=None, version=1.0,
                    framework="ubuntu-sdk-13.10", hooks={}):
        if name is None:
            name = "com.example.%s" % "".join(
                random.choice(string.ascii_lowercase) for i in range(10))
        tmpdir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(tmpdir))
        clickdir = os.path.join(tmpdir, name)
        os.makedirs(clickdir)
        self._create_manifest(os.path.join(clickdir, "manifest.json"),
                              name, version, framework, hooks)
        with open(os.path.join(clickdir, "README"), "w") as f:
            f.write("hello world!")
        with chdir(tmpdir), open(os.devnull, "w") as devnull:
            subprocess.call(
                [self.click_binary, "build", clickdir], stdout=devnull)
        generated_clicks = glob.glob(os.path.join(tmpdir, "*.click"))
        self.assertEqual(len(generated_clicks), 1)
        return generated_clicks[0]
